import pyqtgraph as pg
import numpy as np
from PyQt6.QtCore import pyqtSignal, Qt
from PyQt6.QtWidgets import QVBoxLayout, QSlider, QLabel, QSpacerItem, QSizePolicy, QMessageBox, QDialog, QPushButton
from scipy import signal
from roibaview.gui import BrowseFileDialog
import pandas as pd
from roibaview.plugins.base import BasePlugin


class PeakDetectionPlugin(BasePlugin):
    name = "Peak Detection"
    category = "tool"

    def __init__(self, config=None, parent=None):
        self.config = config
        self.parent = parent
        self.roi_signal = getattr(parent, "signal_roi_idx_changed", None)  # Optional

    def apply(self, *_):
        try:
            # Get controller-level references from the GUI
            controller = getattr(self.parent, "controller", None)
            if not controller:
                raise RuntimeError("Controller not found in GUI context.")

            if not controller.selected_data_sets:
                raise RuntimeError("No dataset selected.")

            # Get current dataset and metadata
            name = controller.selected_data_sets[0]
            dtype = controller.selected_data_sets_type[0]
            roi = controller.current_roi_idx

            data = controller.data_handler.get_data_set(dtype, name)
            meta = controller.data_handler.get_data_set_meta_data(dtype, name)
            fr = meta["sampling_rate"]

            # Get the main data plot
            master_plot = controller.data_plotter.master_plot

            # Create the peak detection dialog
            dialog = PeakDetection(
                data=data + meta["y_offset"],  # match old behavior
                fr=fr,
                master_plot=master_plot,
                roi=roi,
                parent=self.parent
            )

            # Sync with ROI selection signal
            if self.roi_signal:
                self.roi_signal.connect(dialog.roi_changed)
                dialog.finished.connect(lambda: self.roi_signal.disconnect(dialog.roi_changed))

            dialog.exec()

        except Exception as e:
            from PyQt6.QtWidgets import QMessageBox
            QMessageBox.critical(self.parent, "Peak Detection Error", str(e))


# class PeakDetection(QWidget):
class PeakDetection(QDialog):

    signal_roi_changed = pyqtSignal(int)
    main_window_closing = pyqtSignal()

    def __init__(self, data, fr, master_plot, roi, parent=None, **kwargs):
        super().__init__(parent)

        self.roi_idx = roi
        self.data = data  # this is the data set
        self.data_trace = self.data[:, self.roi_idx]  # this is the roi trace
        self.fr = fr
        self.time_axis = self.compute_time_axis(self.data_trace.shape[0], self.fr)
        self.master_plot = master_plot
        self.parameters = dict()
        self.parameters_range = dict()
        self.peaks = dict()
        self.set_parameters()
        self.min_range = 0
        self.max_range = 1000
        # self.set_parameters_range()
        self._init_ui()
        self.main_window_running = True
        self.signal_roi_changed.connect(self.roi_changed)
        self.main_window_closing.connect(self.main_window_is_closing)

    def export_peaks(self):
        file_browser = BrowseFileDialog(self)
        file_dir = file_browser.save_file_name('csv file, (*.csv)')
        if file_dir:
            result = pd.DataFrame()
            result['Time'] = self.peaks['times']
            result['ID'] = self.peaks['idx']
            for k in self.peaks['props']:
                result[k] = self.peaks['props'][k]
            result.to_csv(file_dir)

    @staticmethod
    def compute_time_axis(data_size, fr):
        max_time = data_size / fr
        return np.linspace(0, max_time, data_size)

    def main_window_is_closing(self):
        self.main_window_running = False
        self.close()

    def roi_changed(self, new_roi_idx):
        print('ROI CHANGED')
        self.roi_idx = new_roi_idx
        self.data_trace = self.data[:, self.roi_idx]
        self.peaks['times'], self.peaks['idx'], self.peaks['props'] = self.find_peaks(parameters=self.parameters)
        self.update_plot()

    def _init_ui(self):
        layout = QVBoxLayout()
        self.parameters_labels = dict()
        for param_name in self.parameters.keys():
            label = QLabel(param_name)
            slider = QSlider(Qt.Orientation.Horizontal)
            slider.setMinimum(self.min_range)
            slider.setMaximum(self.max_range)
            slider.setValue(0)  # Initial value
            slider.setTickInterval(1)
            slider.setTickPosition(QSlider.TickPosition.TicksBelow)
            slider.setTickPosition(QSlider.TickPosition.NoTicks)

            self.parameters_labels[param_name] = QLabel('0')

            # Connect slider signal to update function
            slider.valueChanged.connect(lambda value, param=param_name: self.update_parameter(param, value))

            layout.addWidget(label)
            layout.addWidget(slider)
            layout.addWidget(self.parameters_labels[param_name])
            # Add spacer (width, height)
            spacer = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
            layout.addItem(spacer)

        self.export_button = QPushButton('Export...')
        self.export_button.clicked.connect(self.export_peaks)
        layout.addWidget(self.export_button)

        self.setLayout(layout)
        self.setWindowTitle("Peak Detection Parameters")
        # self.show()

    def update_parameter(self, param_name, value):
        # Update parameter value
        if value == 0:
            self.parameters[param_name] = None
        else:
            self.parameters[param_name] = value

        parameters = self.parameters
        if value is not None:
            if param_name == 'height':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 0, np.max(self.data_trace))
            if param_name == 'threshold':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 0, np.max(self.data_trace)*0.1)
            if param_name == 'prominence':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 0, np.max(self.data_trace)*0.25)
            if param_name == 'distance':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 1, self.data_trace.shape[0]*0.5)
            if param_name == 'width':
                parameters[param_name] = self.map_range(value, self.min_range, self.max_range, 1, self.data_trace.shape[0]*0.05)
        else:
            parameters[param_name] = None

        # Update Value Label
        self.parameters_labels[param_name].setText(f'{parameters[param_name]:.2f}')

        # Update find peaks
        self.peaks['times'], self.peaks['idx'], self.peaks['props'] = self.find_peaks(parameters)

        # Update Plot
        self.update_plot()

    def clear_plot(self):
        # check if there is already roi data plotted and remove it
        item_list = self.master_plot.items.copy()
        for item in item_list:
            if item.name() is not None:
                if item.name().startswith('peaks'):
                    self.master_plot.removeItem(item)

    def update_plot(self):
        # Check if there is already roi data plotted and remove it
        self.clear_plot()

        # Create new  plot item
        plot_data_item = pg.ScatterPlotItem(
            # self.peaks['times'], self.data_trace[self.peaks['idx']],
            self.time_axis[self.peaks['idx']], self.data_trace[self.peaks['idx']],
            pen=pg.mkPen(color=(255, 0, 0)),
            brush=pg.mkBrush(color=(255, 0, 0)),
            size=50,
            symbol='arrow_down',
            name=f'peaks',
            skipFiniteCheck=True,
            tip=None,
            hoverable=True,
            hoverSize=100
        )
        # Add plot item to the plot widget
        self.master_plot.addItem(plot_data_item)

    def set_parameters(self):
        self.parameters['height'] = None
        self.parameters['threshold'] = None
        self.parameters['distance'] = None
        self.parameters['prominence'] = None
        # self.parameters['prominence_wlen'] = None
        self.parameters['width'] = None
        # self.parameters['width_rel_height'] = None
        # self.parameters['plateau_size'] = None

    @staticmethod
    def map_range(value, from_min, from_max, to_min, to_max):
        # First, scale the value from the input range to a 0-1 range
        scaled_value = (value - from_min) / (from_max - from_min)

        # Then, scale the value to fit within the output range
        mapped_value = to_min + (to_max - to_min) * scaled_value

        return mapped_value

    def find_peaks(self, parameters):
        peaks, props = signal.find_peaks(self.data_trace, **parameters)
        peaks_time = peaks / self.fr
        return peaks_time, peaks, props

    def closeEvent(self, event):
        if self.main_window_running:
            reply = QMessageBox.question(
                self, 'Message',
                "Are you sure to quit?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No
            )

            if reply == QMessageBox.StandardButton.Yes:
                self._cleanup_on_close()
                self.done(QDialog.DialogCode.Accepted)
                event.accept()
            else:
                event.ignore()
        else:
            self._cleanup_on_close()
            self.done(QDialog.DialogCode.Accepted)
            event.accept()

    def _cleanup_on_close(self):
        """Disconnects signals and clears peak visuals."""
        try:
            self.signal_roi_changed.disconnect(self.roi_changed)
        except Exception:
            pass

        try:
            self.main_window_closing.disconnect()
        except Exception:
            pass

        self.clear_plot()


from roibaview.plugins.base import BasePlugin
from roibaview.gui_utils import DynamicInputDialog
from PyQt6.QtWidgets import QDialog
import pandas as pd
import numpy as np
from scipy.signal import decimate


class ZScore(BasePlugin):
    name = "Z Score"
    category = 'transformation'

    def __init__(self, **kwargs):
        pass

    def apply(self, data, sampling_rate):
        # Check if there is only one ROI (Column)
        if data.shape[1] == 1:
            data = data.flatten()
        return (data - np.mean(data, axis=0)) / np.std(data, axis=0)


class MinMaxTransformation(BasePlugin):
    name = "MinMax Transformation"
    category = 'transformation'

    def __init__(self, **kwargs):
        pass

    def apply(self, data, sampling_rate):
        """ Compute min-max normalization to the range [0, 1]

        :param data: numpy array (columns: ROIs, rows: data points over time)
        :return: Data normalize to the range [0, 1]
        """
        return (data - np.min(data, axis=0)) / (np.max(data, axis=0) - np.min(data, axis=0))


class DeltaFOverF(BasePlugin):
    """ Compute delta F over F for raw fluorescence values
    fbs_per: percentile to calculate baseline (0.0 to 1.0)
    window: window size in seconds for computing sliding percentile baseline (if None, no window is used)
    """

    name = "Delta F over F"
    category = 'transformation'

    def __init__(self, **kwargs):
        pass

    def apply(self, data, sampling_rate):
        fields = {
            'window': (100, 'int'),
            'fbs_per': (10, 'int'),
        }

        dialog = DynamicInputDialog(title="Delta F over F", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()

            # This is using the pandas rolling method, so we need to convert the data to a DataFrame first
            df = pd.DataFrame(data)
            if params['window'] is None:
                fbs = np.percentile(df, params['fbs_per'], axis=0)
            else:
                per_window = int(params['window'] * sampling_rate)
                quant = params['fbs_per'] / 100
                fbs = df.rolling(window=per_window, center=True, min_periods=0).quantile(quant)

            df_over_f = (df - fbs) / fbs
            return df_over_f


class DownSampling(BasePlugin):
    name = "Down Sampling"
    category = 'transformation'

    def __init__(self, **kwargs):
        pass

    def apply(self, data, sampling_rate):
        fields = {
            'ds_factor': (100, 'int'),
        }

        dialog = DynamicInputDialog(title="Down Sampling", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()

            # New sampling rate
            new_fs = sampling_rate / params['ds_factor']

            # Apply decimate function to downsample the data
            down_sampled_data = decimate(data, params['ds_factor'], axis=0)

            return down_sampled_data, new_fs




import cv2
import zipfile
import tifffile
from PyQt6 import QtGui
from PyQt6.QtCore import Qt, QTimer, pyqtSignal, QUrl
from PyQt6.QtWidgets import QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QLabel, QSlider, QPushButton, QFileDialog, QGraphicsItemGroup
import pyqtgraph as pg
from pyqtgraph import ImageView
from roibaview.gui import SimpleInputDialog
from roifile import ImagejRoi
import numpy as np


class ClickableScatterPlotItem(pg.ScatterPlotItem):
    HoverEventSignal = pyqtSignal()

    def __init__(self, *args, roi_index=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.roi_index = roi_index  # Store the ROI index for identification
        self.default_brush = self.opts['brush']  # Save the default brush color
        self.hover_brush = pg.mkBrush(255, 255, 0)  # Yellow for hover
        self.clicked_brush = pg.mkBrush(255, 0, 0)  # Red for click
        self.is_hovered = False

    def hoverEvent(self, ev):
        if ev.isExit():
            # Restore the default color when the mouse leaves the ROI
            self.setBrush(self.default_brush)
            self.is_hovered = False
        elif ev.isEnter():
            # Change to hover color when the mouse enters the ROI
            self.setBrush(self.hover_brush)
            self.is_hovered = True
            self.HoverEventSignal.emit()

    def mouseClickEvent(self, ev):
        if ev.button() == Qt.MouseButton.LeftButton:  # Only handle left button clicks
            if not self.is_hovered:
                # Change to clicked color when the ROI is clicked
                self.setBrush(self.clicked_brush)
                print(f"Clicked on ROI {self.roi_index + 1}")
            else:
                # If it's already hovered, restore the default color
                self.setBrush(self.default_brush)
                self.is_hovered = False


class VideoViewer(QMainWindow):

    FrameChanged = pyqtSignal()
    VideoLoaded = pyqtSignal()
    TimePoint = pyqtSignal(float)
    ConnectToDataTrace = pyqtSignal(bool)

    def __init__(self):
        super().__init__()

        self.video_file = None
        self.current_frame = 0
        self.time_offset = 0

        self.setWindowTitle("Video Viewer")
        self.setGeometry(100, 100, 800, 600)

        # Create the main widget and layout
        self.central_widget = QWidget(self)
        self.setCentralWidget(self.central_widget)
        self.layout = QVBoxLayout(self.central_widget)

        # Create the video frame viewer
        self.info_labels_layout = QVBoxLayout()
        self.video_label = QLabel(f'Please Open A Video File')
        self.video_info_label = QLabel(f'Time Offset: {self.time_offset} s')

        self.info_labels_layout.addWidget(self.video_label)
        self.info_labels_layout.addWidget(self.video_info_label)

        self.image_view = ImageView(self)

        # Connect Mouse Click
        # self.image_view.scene.sigMouseClicked.connect(self.mouse_clicked)

        # self.layout.addWidget(self.video_label)
        # self.layout.addWidget(self.video_info_label)
        self.layout.addLayout(self.info_labels_layout)
        self.layout.addWidget(self.image_view)

        # Create the control widgets
        self.controls_layout = QVBoxLayout()
        self.control_button_layout = QHBoxLayout()

        # Buttons
        self.open_button = QPushButton("Open Video", self)
        self.open_button.clicked.connect(self.open_file_dialog)
        self.control_button_layout.addWidget(self.open_button)

        self.play_button = QPushButton("Play", self)
        self.play_button.clicked.connect(self.play_video)
        self.control_button_layout.addWidget(self.play_button)

        self.pause_button = QPushButton("Pause", self)
        self.pause_button.clicked.connect(self.pause_video)
        self.control_button_layout.addWidget(self.pause_button)

        self.stop_button = QPushButton("Stop", self)
        self.stop_button.clicked.connect(self.stop_video)
        self.control_button_layout.addWidget(self.stop_button)

        self.faster_button = QPushButton("Faster", self)
        self.faster_button.clicked.connect(self.speed_up)
        self.control_button_layout.addWidget(self.faster_button)

        self.slower_button = QPushButton("Slower", self)
        self.slower_button.clicked.connect(self.slow_down)
        self.control_button_layout.addWidget(self.slower_button)

        self.time_offset_button = QPushButton("Time Offset", self)
        self.time_offset_button.clicked.connect(self.add_time_offset)
        self.control_button_layout.addWidget(self.time_offset_button)

        self.rotate_button = QPushButton("Rotate", self)
        self.rotate_button.clicked.connect(self.rotate_video)
        self.control_button_layout.addWidget(self.rotate_button)

        self.connect_video_to_data_trace_button = QPushButton("Connect to Data", self)
        self.connect_video_to_data_trace_button.clicked.connect(self.connect_to_data_trace)
        self.control_button_layout.addWidget(self.connect_video_to_data_trace_button)

        self.load_roi_button = QPushButton("Load ROIs", self)
        self.load_roi_button.clicked.connect(self.open_roi_file_dialog)
        self.control_button_layout.addWidget(self.load_roi_button)

        # Slider
        self.frame_slider = QSlider(Qt.Orientation.Horizontal, self)
        # self.frame_slider.setTickPosition(QSlider.TickPosition.TicksBothSides)
        self.frame_slider.setTickPosition(QSlider.TickPosition.NoTicks)
        self.frame_slider.setTickInterval(1)
        self.frame_slider.valueChanged.connect(self.change_frame)
        self.control_button_layout.addWidget(self.frame_slider)

        self.current_frame_label = QLabel("Current Frame: 0", self)
        self.control_button_layout.addWidget(self.current_frame_label)

        self.current_speed_label = QLabel(", speed lvl: 3", self)
        self.control_button_layout.addWidget(self.current_speed_label)

        self.controls_layout.addWidget(self.frame_slider)
        self.controls_layout.addLayout(self.control_button_layout)
        self.layout.addLayout(self.controls_layout)

        # Disabled Buttons for start up
        self.set_button_state(True)
        self.roi_circle = None

        self.rois = None

        # Initialize video variables
        self._reset_video_viewer()

    def open_roi_file_dialog(self):
        """Open a file dialog to select and load ROI files."""
        roi_file, _ = QFileDialog.getOpenFileName(
            self, "Select ROI File", "", "ROI Files (*.roi *.zip)"
        )
        if roi_file:
            self.rois = self.load_rois(roi_file)

    def load_rois(self, roi_zip_path):
        """Load ROIs from an ImageJ ROI zip file."""
        rois = []
        with zipfile.ZipFile(roi_zip_path, 'r') as zf:
            for filename in zf.namelist():
                if filename.endswith('.roi'):
                    with zf.open(filename) as roi_file:
                        # Read the ROI data as bytes
                        roi_data = roi_file.read()
                        # Create an ROI object from bytes
                        roi = ImagejRoi.frombytes(roi_data)
                        rois.append(roi)
        return rois

    def overlay_rois(self, roi_color, font_size):
        """Overlay ROIs and their labels on the image."""
        for index, roi in enumerate(self.rois):
            # Get ROI coordinates
            coordinates = np.array(roi.coordinates())
            if len(coordinates) == 0:
                continue

            # Draw ROI as a scatter plot and make it clickable and hoverable
            rois_line_plot = pg.PlotDataItem(coordinates[:, 0], coordinates[:, 1], pen=pg.mkPen(255, 255, 255, width=2))
            self.image_view.addItem(rois_line_plot)

            scatter = ClickableScatterPlotItem(pos=coordinates, pen=None, brush=pg.mkBrush(roi_color), size=5,
                                               roi_index=index)
            self.image_view.addItem(scatter)



        for index, roi in enumerate(self.rois):
            # Get ROI coordinates
            coordinates = np.array(roi.coordinates())
            if len(coordinates) == 0:
                continue
            # Add a label next to the ROI
            label_text = f"{index + 1}"  # Example: "ROI 1", "ROI 2", etc.
            label_position = np.median(coordinates, axis=0)
            text_item = pg.TextItem(text=label_text, color=(255, 255, 255))

            # Set the font size
            font = QtGui.QFont()
            font.setPointSize(font_size)
            text_item.setFont(font)

            # Position the label with a slight offset
            text_item.setPos(label_position[0] - 5, label_position[1] - 5)
            self.image_view.addItem(text_item)

    def _reset_video_viewer(self):
        self.image_view.clear()
        self.set_button_state(True)

        # Initialize video variables
        self.video_file = ""
        self.video_frame = None
        self.video_frame_rate = None
        self.current_frame = 0
        self.total_frames = 0
        self.captured_video = None
        self.ms_per_frame = [1, 2, 3, 5, 10, 15, 30, 60, 90, 120, 150, 200, 500, 1000]
        self.ms_per_frame_base = 30
        self.ms_per_frame_id = 6
        self.is_tiff = False
        self.connected_to_data_trace = False
        self.fps = None

        # Create a timer to update the video frames
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_frame)

        # Initialize rotation variables
        self.rotation_angle = 0

    def add_time_offset(self):
        dialog = SimpleInputDialog('Settings', 'Please enter Time Offset [s]: ')
        if dialog.exec() == dialog.DialogCode.Accepted:
            self.time_offset = float(dialog.get_input())
            self.video_info_label.setText(f'Time Offset: {self.time_offset} s')
        else:
            return None

    def mouse_clicked(self, event):
        self.image_view.scene.setClickRadius(20)
        vb = self.image_view.getView()
        scene_coords = event.scenePos()
        key_modifier = event.modifiers()
        # if the click is inside the bounding box of the plot
        if vb.boundingRect().contains(scene_coords):
            mouse_point = vb.mapSceneToView(scene_coords)
            mx = mouse_point.x()
            my = mouse_point.y()
            print(mx, my)
            # self.roi_circle = (int(mx), int(my))
            # self.update_frame()

    def open_file_dialog(self):
        input_file, _ = QFileDialog.getOpenFileName(
            self, "Select Input File", "", "Video Files (*.mp4; *.avi; *.mkv; *.mpg; *.mpeg; *.tif; *.tiff *.TIF; *.TIFF)")
        if input_file:
            self.load_video(input_file)

    def load_video(self, video_file):
        if self.captured_video is not None:
            self.close_file()
            self._reset_video_viewer()

        self.video_file = video_file
        self.current_frame = 0
        self.total_frames = 0

        # Ask for sampling rate
        dialog = SimpleInputDialog('Settings', 'Please enter sampling rate [Hz]: ')
        # if dialog.exec() == QDialog.DialogCode.Accepted:
        if dialog.exec() == dialog.DialogCode.Accepted:
            self.fps = float(dialog.get_input())
        else:
            return None

        if self.video_file.endswith(('.tif', '.tiff', '.TIF', '.TIFF')):
            # This is a tiff file

            # Open the TIFF file in a memory-mapped mode
            self.captured_video = tifffile.TiffFile(video_file, mode='r')
            # Get the number of pages (image stack size)
            self.total_frames = len(self.captured_video.pages)
            self.is_tiff = True
            # Read one frame
            self.video_frame = self.captured_video.pages.get(0).asarray()
        else:
            # This is a video file
            # OpenCV Method:
            self.captured_video = cv2.VideoCapture(self.video_file)

            # This is the frame rate in the video files meta data. This can be wrong!
            # self.fps = self.captured_video.get(cv2.CAP_PROP_FPS)
            self.total_frames = int(self.captured_video.get(cv2.CAP_PROP_FRAME_COUNT))

            self.captured_video.set(cv2.CAP_PROP_POS_FRAMES, self.current_frame - 1)
            ret, self.video_frame = self.captured_video.read()
            self.is_tiff = False

        self.frame_slider.setRange(0, self.total_frames - 1)
        self.current_frame_label.setText(f"Current Frame: {self.current_frame}")
        self.image_view.setImage(self.rotate_frame(self.video_frame), autoLevels=True)
        self.frame_slider.setValue(0)
        self.video_label.setText(f'Sampling Rate: {self.fps:.2f} Hz')

        # Activate Buttons (False: Show Buttons)
        self.set_button_state(False)
        self.VideoLoaded.emit()
        self.connected_to_data_trace = False
        self.connect_video_to_data_trace_button.setText("Connect to Data")

    def set_button_state(self, state):
        self.frame_slider.setDisabled(state)
        self.play_button.setDisabled(state)
        self.pause_button.setDisabled(state)
        self.stop_button.setDisabled(state)
        self.rotate_button.setDisabled(state)
        self.faster_button.setDisabled(state)
        self.slower_button.setDisabled(state)
        self.connect_video_to_data_trace_button.setDisabled(state)

    def play_video(self):
        if self.video_frame is None:
            return
        # ms = 33
        # self.timer.start(ms)  # 30 frames per second (33 milliseconds per frame)
        # ms = self.ms_per_frame[self.ms_per_frame_id]
        ms = int((1 / self.fps)*1000)
        print(ms, end='\r')
        self.timer.start(ms)  # 30 frames per second (33 milliseconds per frame)

    def pause_video(self):
        self.timer.stop()

    def stop_video(self):
        self.timer.stop()
        self.current_frame = 0
        if self.captured_video is not None:
            if self.is_tiff:
                self.video_frame = self.captured_video.pages.get(self.current_frame).asarray()
            else:
                self.captured_video.set(cv2.CAP_PROP_POS_FRAMES, self.current_frame - 1)
                ret, self.video_frame = self.captured_video.read()

            self.image_view.setImage(self.rotate_frame(self.video_frame), autoLevels=True)
            self.current_frame_label.setText(f"Current Frame: {self.current_frame}")
            self.frame_slider.setValue(0)

    def speed_up(self):
        if self.ms_per_frame_id > 0:
            self.ms_per_frame_id -= 1
            self.current_speed_label.setText(f', speed lvl: {self.ms_per_frame_id}')
            self.timer.stop()
            self.play_video()

    def slow_down(self):
        if self.ms_per_frame_id < len(self.ms_per_frame) - 1:
            self.ms_per_frame_id += 1
            self.current_speed_label.setText(f', speed lvl: {self.ms_per_frame_id}')
            self.timer.stop()
            self.play_video()

    def update_frame(self):
        # Update frame while playing video
        if self.captured_video is not None:
            self.current_frame += 1
            if self.current_frame >= self.total_frames:
                self.current_frame = 0
            self.change_frame(self.current_frame)
            # self.FrameChanged.emit()
            self.frame_slider.setValue(self.current_frame)

    def change_frame(self, frame):
        if self.captured_video is not None:
            self.current_frame = frame
            if self.current_frame >= self.total_frames:
                self.current_frame = 0

            self.FrameChanged.emit()
            if self.is_tiff:
                self.video_frame = self.captured_video.pages.get(self.current_frame).asarray()
            else:
                # Capture the next frame
                self.captured_video.set(cv2.CAP_PROP_POS_FRAMES, self.current_frame - 1)
                ret, self.video_frame = self.captured_video.read()
                if not ret:
                    return

            current_time_point = (self.current_frame/self.fps) + self.time_offset
            self.current_frame_label.setText(f"Current Frame / Time: {self.current_frame} / {current_time_point:.2f}")
            self.image_view.setImage(self.rotate_frame(self.video_frame), autoLevels=False)
            # Draw the ROIs on the new frame
            self.TimePoint.emit(current_time_point)

            if self.rois is not None:
                # Overlay ROIs in the desired color
                self.overlay_rois(roi_color=(0, 255, 0), font_size=16)
                self.rois = None

    def connect_to_data_trace(self):
        if not self.connected_to_data_trace:
            self.connected_to_data_trace = True
            self.connect_video_to_data_trace_button.setText("Disconnect")
        else:
            self.connected_to_data_trace = False
            self.connect_video_to_data_trace_button.setText("Connect to Data")

    def rotate_frame(self, frame):
        if self.rotation_angle == 90:
            return cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
        elif self.rotation_angle == 180:
            return cv2.rotate(frame, cv2.ROTATE_180)
        elif self.rotation_angle == 270:
            return cv2.rotate(frame, cv2.ROTATE_90_COUNTERCLOCKWISE)
        else:
            return frame

    def rotate_video(self):
        if self.captured_video is not None:
            self.rotation_angle = (self.rotation_angle + 90) % 360
            self.change_frame(self.current_frame)

    def close_file(self):
        if self.is_tiff:
            # Close the TIFF file
            self.captured_video.close()
        else:
            self.captured_video.release()

    def closeEvent(self, event):
        if self.captured_video is not None:
            self.close_file()
            self._reset_video_viewer()



from PyQt6.QtWidgets import (
    QDialog, QVBoxLayout, QLabel, QLineEdit, QPushButton,
    QDialogButtonBox, QFileDialog, QHBoxLayout, QCheckBox
)
from PyQt6.QtGui import QIntValidator, QDoubleValidator


class DynamicInputDialog(QDialog):
    def __init__(self, title='Input Dialog', fields=None, parent=None):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.fields = fields or {}
        self.inputs = {}  # key -> (widget, type_str)

        layout = QVBoxLayout()

        for label, (default, type_str) in self.fields.items():
            if type_str == 'bool':
                # Checkbox for boolean input
                checkbox = QCheckBox(label)
                checkbox.setChecked(bool(default))
                layout.addWidget(checkbox)
                self.inputs[label] = (checkbox, type_str)

            elif type_str in ('file', 'dir'):
                # File or directory picker
                layout.addWidget(QLabel(label))
                container = QHBoxLayout()
                line_edit = QLineEdit(str(default))
                button = QPushButton("Browse...")
                container.addWidget(line_edit)
                container.addWidget(button)
                layout.addLayout(container)

                def open_dialog(_, le=line_edit, t=type_str):
                    if t == 'file':
                        file_path, _ = QFileDialog.getOpenFileName(self, "Select File")
                        if file_path:
                            le.setText(file_path)
                    else:
                        dir_path, _ = QFileDialog.getSaveFileName(self, "Select Directory", filter="CSV Files (*.csv)")
                        if dir_path:
                            if not dir_path.endswith('.csv'):
                                dir_path += '.csv'
                            le.setText(dir_path)

                button.clicked.connect(open_dialog)
                self.inputs[label] = (line_edit, type_str)

            else:
                # Regular text/numeric input
                layout.addWidget(QLabel(label))
                line_edit = QLineEdit(str(default))
                if type_str == 'int':
                    line_edit.setValidator(QIntValidator())
                elif type_str == 'float':
                    line_edit.setValidator(QDoubleValidator())
                layout.addWidget(line_edit)
                self.inputs[label] = (line_edit, type_str)

        self.add_ok_cancel_buttons(layout)
        self.setLayout(layout)

    def add_ok_cancel_buttons(self, layout):
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        buttonBox = QDialogButtonBox(QBtn)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        layout.addWidget(buttonBox)

    def get_inputs(self):
        result = {}
        for label, (widget, type_str) in self.inputs.items():
            if type_str == 'bool':
                result[label] = widget.isChecked()
            else:
                value = widget.text()
                try:
                    if type_str == 'int':
                        result[label] = int(value)
                    elif type_str == 'float':
                        result[label] = float(value)
                    else:
                        result[label] = value
                except ValueError:
                    result[label] = None
        return result



import sys
from PyQt6.QtWidgets import QApplication
from roibaview.gui import MainWindow
from roibaview.controller import Controller


def main():
    print("RoiBaView is starting...")
    # Start Qt Application
    app = QApplication(sys.argv)
    screen = app.primaryScreen().availableGeometry()

    # GUI
    window = MainWindow(screen)

    # Start Controller
    Controller(gui=window)
    window.show()
    app.exec()


if __name__ == '__main__':
    main()


# import sys
# from PyQt6.QtWidgets import QApplication

# print("Top-level import: before MainWindow")
#
# try:
#     from roibaview.gui import MainWindow
#     print("Imported MainWindow")
# except Exception as e:
#     print("Failed to import MainWindow:", e)
#
# try:
#     from roibaview.controller import Controller
#     print("Imported Controller")
# except Exception as e:
#     print("Failed to import Controller:", e)
#
# def main():
#     print("RoiBaView is starting...")
#     app = QApplication(sys.argv)
#     screen = app.primaryScreen().availableGeometry()
#     window = MainWindow(screen)
#     Controller(gui=window)
#     window.show()
#     app.exec()


import pandas as pd
import os
import tifftools
import shutil
import configparser


class Registrator:
    def __init__(self):
        pass

    @staticmethod
    def start_registration(tiff_file):
        try:
            import suite2p
        except ModuleNotFoundError:
            print('COULD NOT FIND SUITE2P PACKAGE')
            return False

        try:
            suite2p_settings = pd.read_csv('roibaview/suite2p_settings.csv')
            config = configparser.ConfigParser()
            config.read('roibaview/settings.ini')
            # config.sections()
            # config.options('SectionOne')
            # config.get('SectionOne', 'Status')
        except FileNotFoundError:
            print('COULD NOT FIND SUITE2P SETTINGS')
            return False
        embed()
        exit()
        print(f'-- START REGISTRATION --')
        print(f'{tiff_file}')
        print('')
        tiff_file_name = os.path.split(tiff_file)[1]
        tiff_file_dir = os.path.split(tiff_file)[0]

        # SUITE2P SETTINGS
        reg_suite2_path = f'{tiff_file_dir}/suite2p/plane0/reg_tif/'
        ops = suite2p.default_ops()  # populates ops with the default options
        ops['tau'] = suite2p_settings.loc['tau']
        ops['fs'] = suite2p_settings.loc['fs']
        ops['nimg_init'] = suite2p_settings.loc['nimg_init']  # (int, default: 200) how many frames to use to compute reference image for registration
        ops['batch_size'] = suite2p_settings.loc['batch_size']  # (int, default: 200) how many frames to register simultaneously in each batch.
        # Depends on memory constraints - it will be faster to run if the batch is larger, but it will require more RAM.

        ops['reg_tif'] = True  # store reg movie as tiff file
        ops['nonrigid'] = suite2p_settings.loc['nonrigid']  # (bool, default: True) whether or not to perform non-rigid registration,
        # which splits the field of view into blocks and computes registration offset in each block separately.

        ops['block_size'] = [suite2p_settings.loc['block_size'], suite2p_settings.loc['block_size']]  # (two ints, default: [128,128]) size of blocks for non-rigid reg, in pixels.
        # HIGHLY recommend keeping this a power of 2 and/or 3 (e.g. 128, 256, 384, etc) for efficient fft

        ops['roidetect'] = False  # (bool, default: True) whether or not to run ROI detect and extraction

        db = {
            'data_path': [tiff_file_dir],
            'save_path0': tiff_file_dir,
            'tiff_list': [tiff_file_name],
            'subfolders': [],
            'fast_disk': tiff_file_dir,
            'look_one_level_down': False,
        }

        # Run suite2p pipeline in terminal with the above settings
        output_ops = suite2p.run_s2p(ops=ops, db=db)

        print('---------- REGISTRATION FINISHED ----------')
        print('---------- COMBINING TIFF FILES ----------')

        # Load registered tiff files
        f_list = sorted(os.listdir(reg_suite2_path))
        # print('FOUND REGISTERED SINGLE TIFF FILES:')
        # print(f_list)
        # Load first tiff file
        im_combined = tifftools.read_tiff(f'{reg_suite2_path}{f_list[0]}')

        # Combine tiff files to one file
        for k, v in enumerate(f_list):
            if k == 0:
                continue
            else:
                im_dummy = tifftools.read_tiff(f'{reg_suite2_path}{v}')
                im_combined['ifds'].extend(im_dummy['ifds'])
        # Store combined tiff file
        tifftools.write_tiff(im_combined, f'{tiff_file_dir}/Registered_{tiff_file_name}')

        # Delete all temporary files
        shutil.rmtree(f'{tiff_file_dir}/suite2p/')

        print('----------------------------------------')
        print('Stored Registered Tiff File to HDD')




import os
import numpy as np
import pandas as pd
import configparser
from datetime import datetime
from PyQt6.QtWidgets import QMessageBox, QListWidget, QListWidgetItem, QDialog, QApplication
from PyQt6.QtCore import pyqtSignal, QObject, Qt
from PyQt6.QtGui import QPen, QBrush, QColor
import pyqtgraph as pg
from roibaview.data_handler import DataHandler, TransformData
from roibaview.gui import BrowseFileDialog, InputDialog, SimpleInputDialog, ChangeStyle, MessageBox
from roibaview.data_plotter import DataPlotter, PyqtgraphSettings
# from roibaview.peak_detection import PeakDetection
# from roibaview.ventral_root_detection import VentralRootDetection
from roibaview.custom_view_box import CustomViewBoxMenu
from roibaview.registration import Registrator
from roibaview.video_viewer import VideoViewer
# from roibaview.video_converter import VideoConverter
from roibaview.plugins.loader import load_plugins


class Controller(QObject):
    # ==================================================================================================================
    # SIGNALS
    # ------------------------------------------------------------------------------------------------------------------
    signal_import_traces = pyqtSignal()
    signal_roi_idx_changed = pyqtSignal()
    signal_closing = pyqtSignal()

    # ==================================================================================================================
    # INITIALIZING
    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, gui):
        QObject.__init__(self)
        # Create GUI
        self.gui = gui
        self.gui.controller = self  # Expose controller to plugins via GUI
        self.gui.closeEvent = self.closeEvent
        self.mouse_x_pos = 0

        # Get a DataHandler
        self.data_handler = DataHandler()
        self.selected_data_sets = []
        self.selected_data_sets_type = []
        self.selected_data_sets_rows = []
        self.selected_data_sets_items = []
        self.current_roi_idx = 0

        # Get a Data Transformer
        self.data_transformer = TransformData()

        # # Replace View Box menu
        # self.view_box = self.gui.trace_plot_item.getViewBox()
        # self.view_box.menu = CustomViewBoxMenu(self.view_box)

        # View Box Right Click Context Menu
        # Hide "Plot Options"
        self.gui.trace_plot_item.ctrlMenu.menuAction().setVisible(False)

        # Get a Video Viewer
        self.video_viewer = VideoViewer()
        self.video_viewers = []
        # self.video_converter = None

        # Get DataPlotter
        self.data_plotter = DataPlotter(self.gui.trace_plot_item)

        # Get a File Browser
        self.file_browser = BrowseFileDialog(self.gui)

        # Get a Peak Detector
        # self.peak_detection = None

        # Get a VR Detector
        self.vr_detection = None

        # Set Selection Mode
        self.signal_selection_status = False
        self.cut_out_region = None

        # Establish Connections to Buttons and Menus
        self.connections()

        # KeyBoard Bindings
        # self.gui.key_pressed.connect(self.on_key_press)
        # self.gui.key_released.connect(self.on_key_release)
        self.gui.trace_plot_item.scene().sigMouseMoved.connect(self.mouse_moved)

        self.pyqtgraph_settings = PyqtgraphSettings()

        # Check Config File
        package_dir = os.path.dirname(__file__)
        config_path = os.path.join(package_dir, 'config.ini')
        self.config_name = config_path

        if not os.path.exists(self.config_name):
            self._create_config_file()
        else:
            self.config = configparser.ConfigParser()
            self.config.read(self.config_name)

        # Load plugins with context (pass config and GUI as needed)
        self.plugins = load_plugins(context={
            'config': self.config,  # or other settings object
            'parent': self.gui  # used for dialog-based plugins
        })

        # Categorize them for use
        self.tool_plugins = [p for p in self.plugins if p.category == "tool"]
        self.utils_plugins = [p for p in self.plugins if p.category == "utils"]

        # Context Menu
        self.filter_plugins = [p for p in self.plugins if p.category == "filter"]
        self.transformation_plugins = [p for p in self.plugins if p.category == "transformation"]

        # Register tools in Tools menu
        self.gui.add_tools_menu_plugins(self.tool_plugins)

        # Register utils in Utils menu
        self.gui.add_utils_menu_plugins(self.utils_plugins)

        # Populate Context Menu Filter Plugins
        self.gui.populate_filter_plugins_menu(self.filter_plugins, self.apply_filter_plugin)
        self.gui.populate_transformation_plugins_menu(self.transformation_plugins, self.apply_filter_plugin)

    def apply_filter_plugin(self, plugin):
        if not self.selected_data_sets:
            return

        for name, kind in zip(self.selected_data_sets, self.selected_data_sets_type):
            data = self.data_handler.get_data_set(kind, name)
            meta = self.data_handler.get_data_set_meta_data(kind, name)
            result = plugin.apply(data, meta['sampling_rate'])
            if type(result) is tuple:  # function returns two values, the second is the new sampling rate (down sampling)
                fr = result[1]
                result = result[0]
            else:
                fr = meta['sampling_rate']
            new_name = f"{name}_{plugin.name.replace(' ', '_')}"
            already_exists = self.data_handler.add_new_data_set(
                data_set_type=kind,
                data_set_name=new_name,
                data=result,
                sampling_rate=fr,
                time_offset=0,
                y_offset=0,
                header=meta.get("roi_names", list(range(result.shape[1])))
            )
            if already_exists:
                data_set_name = new_name + '_new'
                self.add_data_set_to_list(kind, data_set_name)
            else:
                self.add_data_set_to_list(kind, new_name)

    def _create_config_file(self):
        self.config = configparser.ConfigParser()
        self.config['FFMPEG'] = {'dir': 'NaN'}
        with open(self.config_name, 'w') as configfile:
            self.config.write(configfile)

    def connections(self):
        # File Menu
        self.gui.file_menu_import_csv.triggered.connect(self.import_csv_file)
        self.gui.file_menu_new_viewer_file.triggered.connect(self.new_file)
        self.gui.file_menu_save_viewer_file.triggered.connect(self.save_file)
        self.gui.file_men_open_viewer_file.triggered.connect(self.open_file)
        self.gui.file_menu_action_exit.triggered.connect(self.gui.close)

        # DataSets List
        # Connect item selection changed signal
        self.gui.data_sets_list.itemSelectionChanged.connect(self.data_set_selection_changed)
        # self.gui.data_sets_list.itemActivated.connect(lambda: print('CLICK'))

        # Arrow Buttons
        self.gui.next_button.clicked.connect(self.next_roi)
        self.gui.prev_button.clicked.connect(self.prev_roi)

        # ROI changed
        self.signal_roi_idx_changed.connect(lambda: self.update_plots(change_global=False))
        # self.signal_roi_idx_changed.connect(self.check_peak_detector)

        # Context Menu
        self.gui.data_sets_list_rename.triggered.connect(self.rename_data_set)
        self.gui.data_sets_list_delete.triggered.connect(self.delete_data_set)
        self.gui.data_sets_list_delete_col.triggered.connect(self.delete_column)

        self.gui.data_sets_list_export.triggered.connect(self.export_to_csv)
        self.gui.data_sets_list_time_offset.triggered.connect(self.time_offset)
        self.gui.data_sets_list_y_offset.triggered.connect(self.y_offset)

        # Style Submenu
        self.gui.style_color.triggered.connect(self.pick_color)
        self.gui.style_lw.triggered.connect(self.pick_lw)

        # Tools
        # Video Viewer
        self.gui.tools_menu_open_video_viewer.triggered.connect(self.open_video_viewer)
        self.video_viewer.TimePoint.connect(self.connect_video_to_plot)

        # Remove Column from csv file
        self.gui.tools_menu_csv_remove_column.triggered.connect(self.csv_remove_column)

        # Convert Ventral Root
        self.gui.tools_menu_convert_ventral_root.triggered.connect(self.convert_ventral_root)

        # Create Stimulus from File
        self.gui.tools_menu_create_stimulus.triggered.connect(self.create_stimulus_from_file)

        # Ventral Root Event Detection
        # self.gui.tools_menu_detect_vr.triggered.connect(self._ventral_root_detection)

        # KeyBoard Bindings
        self.gui.key_pressed.connect(self.on_key_press)
        self.data_plotter.master_plot.scene().sigMouseClicked.connect(self.on_mouse_click)

    def delete_column(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot rename multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None
        if len(self.selected_data_sets) == 0:
            return None

        # Get Column Nr
        dialog = SimpleInputDialog(title='Settings', text='Please enter some stuff: ')
        if dialog.exec() == QDialog.DialogCode.Accepted:
            received = dialog.get_input()
        else:
            return None

        col_nr = int(received)
        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)

        # Remove the column
        self.data_handler.delete_column(data_set_type, data_set_name, col_nr)

    def draw_selection(self, status='start'):
        if status == 'exit' and self.cut_out_region is not None:
            self.data_plotter.master_plot.removeItem(self.cut_out_region)
        else:
            # Add a LinearRegionItem to select a region
            self.cut_out_region = pg.LinearRegionItem(
                [self.mouse_x_pos, self.mouse_x_pos + 10],
                brush=QBrush(QColor(255, 0, 0, 50)),
                pen=pg.mkPen(color=(255, 0, 0), width=5),
                hoverBrush=QBrush(QColor(255, 0, 0, 100)),
                hoverPen=pg.mkPen(color=(0, 255, 0), width=5),
                movable=True,
                bounds=None,
                swapMode='sort',
                clipItem=None
            )
            self.data_plotter.master_plot.addItem(self.cut_out_region)

    def cut_selection(self):
        # Get the selected region boundaries
        if self.cut_out_region is not None:

            # Get tag info from user
            dialog = SimpleInputDialog('Save Selection', 'Tag:')
            if dialog.exec() == dialog.DialogCode.Accepted:
                tag_name = dialog.get_input()
            else:
                return None

            min_x, max_x = self.cut_out_region.getRegion()

            # Get all data that is visible on the plot
            # Prepare Data Frame for csv file
            results = pd.DataFrame()
            k = 0
            for data_set in self.data_plotter.master_plot.listDataItems():
                column_name = f'{data_set.name()}_{tag_name}'
                x = data_set.getData()[0]
                y = data_set.getData()[1]

                if y is not None and x is not None:
                    # Find the indices of the region
                    min_idx = np.searchsorted(x, min_x)
                    max_idx = np.searchsorted(x, max_x)

                    # Extract and store the selected region
                    if k == 0:
                        selected_x = x[min_idx:max_idx]
                        results['Time'] = selected_x

                    selected_y = y[min_idx:max_idx]
                    results[column_name] = selected_y
                    k += 1

            # Store to HDD
            results.to_csv(f'roibaview/temp/{datetime.now().strftime("%Y_%m_%d_%H_%M_%S")}_{tag_name}.csv')
            # Remove Selection Markers from Plot
            self.data_plotter.master_plot.removeItem(self.cut_out_region)

    # def _ventral_root_detection(self):
    #     if len(self.selected_data_sets) > 0:
    #         self.gui.freeze_gui(True)
    #         data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(k=0)
    #         current_data_set = self.data_handler.get_data_set(data_set_name=data_set_name, data_set_type=data_set_type)
    #         meta_data = self.data_handler.get_data_set_meta_data(data_set_type=data_set_type, data_set_name=data_set_name)
    #
    #         self.vr_detection = VentralRootDetection(
    #             data=current_data_set,
    #             fr=meta_data['sampling_rate'],
    #             master_plot=self.data_plotter.master_plot,
    #             roi=self.current_roi_idx,
    #         )
    #         self.vr_detection.show()
    #         if self.vr_detection.exec() == QDialog.DialogCode.Accepted:
    #             self.gui.freeze_gui(False)
    #             self.vr_detection = None

    def export_to_csv(self):
        file_dir = self.file_browser.save_file_name('csv file, (*.csv *.txt)')
        if file_dir:
            if len(self.selected_data_sets) > 1:
                dlg = QMessageBox()
                dlg.setWindowTitle('ERROR')
                dlg.setText(f'You cannot export multiple data sets at once!')
                button = dlg.exec()
                if button == QMessageBox.StandardButton.Ok:
                    return None
            if len(self.selected_data_sets) == 0:
                return None

            # Export selected data set to csv
            data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
            # meta_data = self.data_handler.get_data_set_meta_data(data_set_type=data_set_type, data_set_name=data_set_name)
            df = pd.DataFrame(self.data_handler.get_data_set(data_set_type=data_set_type, data_set_name=data_set_name))
            df.to_csv(file_dir, index=False, header=False)

    def create_stimulus_from_file(self):
        file_dir = self.file_browser.browse_file('csv file, (*.csv *.txt)')
        if file_dir:
            dialog = InputDialog(dialog_type='stimulus')
            if dialog.exec() == QDialog.DialogCode.Accepted:
                # name
                received = dialog.get_input()
                data_set_name = received['name']
            else:
                return None

            protocol = pd.read_csv(file_dir)
            t_max = max(protocol.max())
            # t0 = 0
            fr = 100
            stimulus = np.zeros(int(fr * t_max))
            for on, off in zip(protocol.iloc[:, 0], protocol.iloc[:, 1]):
                start = int(on * fr)
                end = int(off * fr)
                stimulus[start:end] = 1

            # Prepare it for hdf5 matrix form
            stimulus = list(stimulus[:, np.newaxis])
            data_set_type = 'global_data_sets'
            # Add stimulus to data sets
            self.data_handler.add_new_data_set(
                data_set_type=data_set_type,
                data_set_name=data_set_name,
                data=stimulus,
                sampling_rate=fr,
                time_offset=0,
                y_offset=0,
                header='Stimulus')

            # Add new data set to the list in the GUI
            self.add_data_set_to_list(data_set_type, data_set_name)

    def convert_ventral_root(self):
        """
        Expects following file structure:
        └── vr_data
            ├── sweep_01
            │   ├── sw_01_01.txt
            │   ├── sw_01_02.txt
            │   :
            │   └── sw_01_n.txt
            ├── sweep_02
            :
            └── sweep_n

        All vr text files from one sweep will then be combined into one meaningful and solid data file (csv)
        :return:
        """
        from joblib import Parallel, delayed
        from roibaview.ventral_root import transform_ventral_root_parallel, pickle_stuff
        file_structure = '''
        Expects following file structure:
        └── vr_data
            ├── sweep_01
            │   ├── sw_01_01.txt
            │   ├── sw_01_02.txt
            │   :
            │   └── sw_01_n.txt
            ├── sweep_02
            :
            └── sweep_n
            
        All vr text files from one sweep will then be combined into one meaningful and solid data file (csv)
        '''

        store_to_dict = False
        vr_rec_dur = 60  # in secs
        vr_fr = 10000  # in Hz

        # Get directory
        file_dir = self.file_browser.browse_directory()
        save_dir = file_dir
        if file_dir:
            print('========== VENTRAL ROOT ==========')
            print('')
            print(file_structure)
            print('')
            print(f'Ventral Root Recording Files in: {file_dir}')
            print(f'Store result to: {save_dir}')
            print('++++ START PROCESSING +++')
            print('This can take some time ...')
            print( 'This relies heavily on CPU, RAM and HDD. HDD is normally the bottleneck, so make sure to use a fast one!')
            print('... Please Wait ...')
            print('')
            # Process all sweeps in parallel
            sweep_numbers = os.listdir(file_dir)
            results = Parallel(n_jobs=-2)(delayed(
                transform_ventral_root_parallel)(save_dir, file_dir, vr_rec_dur, vr_fr, i) for i in
                                          sweep_numbers)
            if store_to_dict:
                print('STORING DICT TO HDD')
                vr_recordings = dict()
                for res in results:
                    sw = list(res.keys())[0]
                    vr_recordings[sw] = res[sw]
                pickle_stuff(f'{save_dir}/all_ventral_root.pickle', data=vr_recordings)

            print('++++ FINISHED PROCESSING ++++')

    def pick_lw(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot change color of multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None

        if len(self.selected_data_sets) == 0:
            return None

        lw = ChangeStyle().get_lw()
        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
        self.data_handler.add_meta_data(data_set_type=data_set_type, data_set_name=data_set_name, metadata_dict={'lw': lw})
        self.update_plots(change_global=True)

    def pick_color(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot change color of multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None

        if len(self.selected_data_sets) == 0:
            return None

        color = ChangeStyle().get_color()
        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
        self.data_handler.add_meta_data(data_set_type=data_set_type, data_set_name=data_set_name, metadata_dict={'color': color})
        self.update_plots(change_global=True)

    def csv_remove_column(self):
        csv_handler = CSVHandler(self.gui)
        csv_handler.remove_column_from_csv_file()
        print('YUHUUUU')

    def connect_video_to_plot(self, time_point):
        for v in self.video_viewers:
            if v.connected_to_data_trace:
                if len(self.selected_data_sets) > 0:
                    y_min = []
                    y_max = []
                    for data_set_name, data_set_type in zip(self.selected_data_sets, self.selected_data_sets_type):
                        if data_set_type == 'data_sets':
                            data = self.data_handler.get_roi_data(data_set_name, roi_idx=self.current_roi_idx)
                        else:
                            data = self.data_handler.get_data_set('global_data_sets', data_set_name)
                        y_min.append(np.min(data))
                        y_max.append(np.max(data))
                    y_range = [np.min(y_min), np.max(y_max)]
                    self.data_plotter.update_video_plot(time_point, y_range)
                else:
                    self.data_plotter.clear_plot_data(name='video')
            else:
                self.data_plotter.clear_plot_data(name='video')

    def open_video_viewer(self):
        self.video_viewers.append(VideoViewer())
        self.video_viewers[-1].show()
        self.video_viewers[-1].TimePoint.connect(self.connect_video_to_plot)

        # self.video_viewer = VideoViewer()
        # self.video_viewer.show()

    # def open_video_converter(self):
    #     self.video_converter = VideoConverter(self.config)
    #     self.video_converter.show()

    def y_offset(self):
        if len(self.selected_data_sets) > 0:
            for k, _ in enumerate(self.selected_data_sets):
                data_set_name = self.selected_data_sets[k]
                data_set_type = self.selected_data_sets_type[k]
                meta_data = self.data_handler.get_data_set_meta_data(
                    data_set_type=data_set_type,
                    data_set_name=data_set_name
                )

                # Get settings by user input
                dialog = SimpleInputDialog('Settings', 'Please enter Y Offset: ', default_value=meta_data['y_offset'])
                if dialog.exec() == dialog.DialogCode.Accepted:
                    y_offset = {'y_offset': float(dialog.get_input())}
                    self.data_handler.add_meta_data(data_set_type, data_set_name, y_offset)
                    self.update_plots(change_global=True)
                else:
                    return None

    def time_offset(self):
        if len(self.selected_data_sets) > 0:
            for k, _ in enumerate(self.selected_data_sets):
                data_set_name = self.selected_data_sets[k]
                data_set_type = self.selected_data_sets_type[k]
                meta_data = self.data_handler.get_data_set_meta_data(
                    data_set_type=data_set_type,
                    data_set_name=data_set_name
                )

                # Get settings by user input
                dialog = SimpleInputDialog('Settings', 'Please enter Time Offset [s]: ', default_value=meta_data['time_offset'])
                if dialog.exec() == dialog.DialogCode.Accepted:
                    time_offset = {'time_offset': float(dialog.get_input())}
                    self.data_handler.add_meta_data(data_set_type, data_set_name, time_offset)
                    self.update_plots(change_global=True)
                else:
                    return None

    def rename_data_set(self):
        if len(self.selected_data_sets) > 1:
            dlg = QMessageBox()
            dlg.setWindowTitle('ERROR')
            dlg.setText(f'You cannot rename multiple data sets at once!')
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Ok:
                return None
        if len(self.selected_data_sets) == 0:
            return None

        data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(0)
        # Get settings by user input
        dialog = InputDialog(dialog_type='rename')
        if dialog.exec() == QDialog.DialogCode.Accepted:
            received = dialog.get_input()
            new_name = received['data_set_name']
            if new_name != '':
                self.data_handler.rename_data_set(data_set_type=data_set_type, data_set_name=data_set_name, new_name=new_name)
                # self.remove_selected_data_set_from_list(data_set_name, data_set_item)
                # self.add_data_set_to_list(data_set_type, new_name)
                self.rename_item_from_list(data_set_item=data_set_item, new_name=new_name)
            else:
                dlg = QMessageBox()
                dlg.setWindowTitle('ERROR')
                dlg.setText(f'Please enter a valid name!')
                button = dlg.exec()
                if button == QMessageBox.StandardButton.Ok:
                    return None
        else:
            return None

    def delete_data_set(self):
        if len(self.selected_data_sets) > 0:
            dlg = QMessageBox(self.gui)
            dlg.setWindowTitle('Delete Data Set')
            dlg.setText(f'Are You Sure You Want To Delete The Selected Data Sets" ?')
            dlg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
            dlg.setIcon(QMessageBox.Icon.Question)
            button = dlg.exec()
            if button == QMessageBox.StandardButton.Yes:
                for dt, ds, item in zip(self.selected_data_sets_type, self.selected_data_sets, self.selected_data_sets_items):
                    # print(ds, item)
                    # print('')
                    self.data_handler.delete_data_set(dt, ds)
                    self.remove_selected_data_set_from_list(ds, item)

    def tiff_registration(self):
        registrator = Registrator()
        # Get file dir
        file_dir = self.file_browser.browse_file('tiff file, (*.tiff; *.tif)')
        if file_dir:
            registrator.start_registration(file_dir)

    # def check_peak_detector(self):
    #     if self.peak_detection is not None:
    #         self.peak_detection.roi_changed(self.current_roi_idx)  # This will trigger the signal
    #     if self.vr_detection is not None:
    #         self.vr_detection.roi_changed(self.current_roi_idx)  # This will trigger the signal

    # def _start_peak_detection(self):
    #     if len(self.selected_data_sets) > 0:
    #         self.gui.freeze_gui(True)
    #         data_set_name, data_set_type, data_set_item = self.get_selected_data_sets(k=0)
    #         # current_data = self.data_handler.get_roi_data(data_set_name, roi_idx=self.current_roi_idx)
    #         current_data_set = self.data_handler.get_data_set(data_set_name=data_set_name, data_set_type=data_set_type)
    #         meta_data = self.data_handler.get_data_set_meta_data(data_set_type=data_set_type, data_set_name=data_set_name)
    #
    #         self.peak_detection = PeakDetection(
    #             data=current_data_set + meta_data['y_offset'],
    #             fr=meta_data['sampling_rate'],
    #             master_plot=self.data_plotter.master_plot,
    #             roi=self.current_roi_idx,
    #         )
    #         # self.peak_detection.signal_roi_changed.connect(lambda value: print("Variable changed:", value))
    #         self.peak_detection.show()
    #         if self.peak_detection.exec() == QDialog.DialogCode.Accepted:
    #             self.gui.freeze_gui(False)
    #             self.peak_detection = None
    #         # self.peak_detection.exec()

    def get_selected_data_sets(self, k):
        data_set_name = self.selected_data_sets[k]
        data_set_type = self.selected_data_sets_type[k]
        data_set_item = self.selected_data_sets_items[k]
        return data_set_name, data_set_type, data_set_item

    def filter_data(self, mode):
        if len(self.selected_data_sets) > 0:
            for k, _ in enumerate(self.selected_data_sets):
                data_set_name = self.selected_data_sets[k]
                data_set_type = self.selected_data_sets_type[k]
                data = self.data_handler.get_data_set(data_set_type=data_set_type, data_set_name=data_set_name)
                meta_data = self.data_handler.get_data_set_meta_data(data_set_type=data_set_type, data_set_name=data_set_name)
                fr = meta_data['sampling_rate']
                filtered_data = None
                if mode == 'moving_average':
                    # Get settings by user input
                    dialog = InputDialog(dialog_type='moving_average')
                    if dialog.exec() == QDialog.DialogCode.Accepted:
                        received = dialog.get_input()
                        win = float(received['window'])
                    else:
                        return None
                    filtered_data = self.data_transformer.filter_moving_average(data, fr=fr, window=win)

                if mode == 'diff':
                    filtered_data = self.data_transformer.filter_differentiate(data)

                if mode == 'lowpass':
                    # Get settings by user input
                    dialog = InputDialog(dialog_type='butter')
                    if dialog.exec() == QDialog.DialogCode.Accepted:
                        received = dialog.get_input()
                        cutoff = float(received['cutoff'])
                    else:
                        return None
                    filtered_data = self.data_transformer.filter_low_pass(data, cutoff, fs=fr)

                if mode == 'highpass':
                    # Get settings by user input
                    dialog = InputDialog(dialog_type='butter')
                    if dialog.exec() == QDialog.DialogCode.Accepted:
                        received = dialog.get_input()
                        cutoff = float(received['cutoff'])
                    else:
                        return None
                    filtered_data = self.data_transformer.filter_high_pass(data, cutoff, fs=fr)

                if mode == 'env':
                    # Get settings by user input
                    dialog = InputDialog(dialog_type='butter')
                    if dialog.exec() == QDialog.DialogCode.Accepted:
                        received = dialog.get_input()
                        cutoff = float(received['cutoff'])
                    else:
                        return None
                    filtered_data = self.data_transformer.envelope(data, cutoff, fr)

                if mode == 'ds':
                    # Get settings by user input
                    dialog = InputDialog(dialog_type='ds')
                    if dialog.exec() == QDialog.DialogCode.Accepted:
                        received = dialog.get_input()
                        ds_factor = int(received['ds_factor'])
                    else:
                        return None
                    filtered_data, fr = self.data_transformer.down_sampling(data, ds_factor, fr)

                if filtered_data is not None:
                    # Create a new data set from this
                    check = self.data_handler.add_new_data_set(
                        data_set_type=data_set_type,
                        data_set_name=f'{data_set_name}_{mode}',
                        data=filtered_data,
                        sampling_rate=fr,
                        time_offset=meta_data['time_offset'],
                        y_offset=meta_data['y_offset'],
                    )
                    # Add new data set to the list in the GUI
                    if check:
                        # data name already exists, so we have to change it
                        data_set_name = f'{data_set_name}_{mode}' + '_new'
                    else:
                        data_set_name = f'{data_set_name}_{mode}'
                    self.add_data_set_to_list(data_set_type, data_set_name)

    def context_menu(self, mode):
        result = None
        if len(self.selected_data_sets) > 0:
            for k, _ in enumerate(self.selected_data_sets):
                # k = 0
                data_set_name = self.selected_data_sets[k]
                data_set_type = self.selected_data_sets_type[k]
                # if data_set_type != 'data_sets':
                #     return None
                data = self.data_handler.get_data_set(data_set_type=data_set_type, data_set_name=data_set_name)
                meta_data = self.data_handler.get_data_set_meta_data(data_set_type=data_set_type, data_set_name=data_set_name)
                if mode == 'df_f':
                    # Get settings by user input
                    dialog = InputDialog(dialog_type='df_over_f')
                    if dialog.exec() == QDialog.DialogCode.Accepted:
                        received = dialog.get_input()
                        fbs_per = float(received['fbs_per'])
                        fbs_win = float(received['fbs_window'])
                    else:
                        return None

                    result = self.data_transformer.to_delta_f_over_f(data, fbs_per=fbs_per, fr=meta_data['sampling_rate'], window=fbs_win)

                if mode == 'z':
                    result = self.data_transformer.to_z_score(data)

                if mode == 'min_max':
                    result = self.data_transformer.to_min_max(data)
                    # print('MIN_MAX')

                # Create a new data set from this
                if result is not None:
                    check = self.data_handler.add_new_data_set(
                        data_set_type=data_set_type,
                        data_set_name=f'{data_set_name}_{mode}',
                        data=result,
                        sampling_rate=meta_data['sampling_rate'],
                        time_offset=meta_data['time_offset'],
                        y_offset=meta_data['y_offset'],
                        header=f'{data_set_name}_{mode}'
                    )
                    # Add new data set to the list in the GUI
                    if check:
                        # data set name already exists, so whe have to change it
                        data_set_name = f'{data_set_name}_{mode}' + '_new'
                    else:
                        data_set_name = f'{data_set_name}_{mode}'
                    self.add_data_set_to_list(data_set_type, data_set_name)

    def next_roi(self):
        # First check if there are active data sets
        if 'data_sets' in self.selected_data_sets_type:
            self.current_roi_idx = (self.current_roi_idx + 1) % self.data_handler.roi_count
            self.signal_roi_idx_changed.emit()

    def prev_roi(self):
        # First check if there are active data sets
        if 'data_sets' in self.selected_data_sets_type:
            self.current_roi_idx = (self.current_roi_idx - 1) % self.data_handler.roi_count
            self.signal_roi_idx_changed.emit()

    def update_plots(self, change_global=True):
        # print(f'ROI: {self.current_roi_idx}')
        # get new roi data
        roi_data = []
        time_points = []
        global_data = []
        global_time_points = []
        meta_data_list = list()
        global_meta_data_list = list()

        for data_set_name, data_set_type in zip(self.selected_data_sets, self.selected_data_sets_type):
            if data_set_type == 'data_sets':
                r = self.data_handler.get_roi_data(data_set_name, roi_idx=self.current_roi_idx)
                meta_data = self.data_handler.get_data_set_meta_data('data_sets', data_set_name)
                fr = meta_data['sampling_rate']
                # print(f'{data_set_name}: fr={fr} Hz (shape={r.shape[0]}) samples')
                time_offset = meta_data['time_offset']
                y_offset = meta_data['y_offset']
                try:
                    time_points.append(self.data_handler.compute_time_axis(r.shape[0], fr) + time_offset)
                except AttributeError:
                    from IPython import embed
                    embed()
                    exit()

                roi_data.append(r + y_offset)
                meta_data_list.append(meta_data)
            if data_set_type == 'global_data_sets' and change_global:
                r = self.data_handler.get_data_set('global_data_sets', data_set_name)
                meta_data = self.data_handler.get_data_set_meta_data('global_data_sets', data_set_name)
                fr = meta_data['sampling_rate']
                time_offset = meta_data['time_offset']
                y_offset = meta_data['y_offset']
                global_time_points.append(self.data_handler.compute_time_axis(r.shape[0], fr) + time_offset)
                global_data.append(r + y_offset)
                global_meta_data_list.append(meta_data)

        # Update Plot
        if len(roi_data) > 0:
            self.data_plotter.update(time_points, roi_data, meta_data_list)
            self.data_plotter.master_plot.setTitle(f'ROI: {self.current_roi_idx+1}')
        else:
            self.data_plotter.clear_plot_data(name='data')

        # Update Global Plot
        if change_global:
            if len(global_data) > 0:
                self.data_plotter.update_global(global_time_points, global_data, global_meta_data_list)
            else:
                self.data_plotter.clear_plot_data(name='global')

    def data_set_selection_changed(self):
        # Get selected data sets
        self.selected_data_sets = [item.text() for item in self.gui.sender().selectedItems()]
        self.selected_data_sets_type = [item.data(1) for item in self.gui.sender().selectedItems()]
        # self.selected_data_sets_rows = [k for k in self.gui.sender().selectedItems()]
        self.selected_data_sets_items = [k for k in self.gui.sender().selectedItems()]

        # print('')
        # print(f"Selected Items: {self.selected_data_sets}, QItem: {self.selected_data_sets_items}")
        # print('')

        # Update Plots
        self.update_plots()

    def import_csv_file(self):
        # Get file dir
        file_dir = self.file_browser.browse_file('csv file, (*.csv *.txt)')
        if file_dir:
            # Get data set name by user
            # dialog = ImportCsvDialog()
            dialog = InputDialog(dialog_type='import_csv')
            if dialog.exec() == QDialog.DialogCode.Accepted:
                # data_set_name, fr, is_global = dialog.get_settings()
                received = dialog.get_input()
                data_set_name = received['data_set_name']
                fr = float(received['fr'])
                is_global = received['is_global']
            else:
                return None

            if is_global:
                data_set_type = 'global_data_sets'
            else:
                data_set_type = 'data_sets'

            # check if the data set name already exists
            check_if_exists = self.data_handler.check_if_exists(data_set_type, data_set_name)
            if check_if_exists:
                MessageBox(title='ERROR', text='Data set with this name already exists! Will rename it.')
                data_set_name = data_set_name + '_new'

            # Import csv file
            self.data_handler.import_csv(file_dir=file_dir, data_name=data_set_name, sampling_rate=fr, data_set_type=data_set_type)

            # Add new data set to the list in the GUI
            self.add_data_set_to_list(data_set_type, data_set_name)

    def add_data_set_to_list(self, data_set_type, data_set_name):
        # row = self.gui.data_sets_list.count()
        # Add new data set to the list in the GUI
        new_list_item = QListWidgetItem()
        new_list_item.setText(data_set_name)
        new_list_item.setData(1, data_set_type)
        # new_list_item.setData(3, row)
        self.gui.data_sets_list.addItem(new_list_item)

    def rename_item_from_list(self, data_set_item, new_name):
        item = self.gui.data_sets_list.item(self.gui.data_sets_list.row(data_set_item))
        item.setText(new_name)

    def remove_selected_data_set_from_list(self, data_set_name, data_set_item):
        # list_items = self.gui.data_sets_list.selectedItems()
        # if not list_items:
        #     return
        # for item in list_items:
        #     self.gui.data_sets_list.takeItem(self.gui.data_sets_list.row(item))
        if not data_set_name:
            return
        else:
            self.gui.data_sets_list.takeItem(self.gui.data_sets_list.row(data_set_item))

    def save_file(self):
        file_dir = self.file_browser.save_file_name('hdf5 file, (*.hdf5)')
        if file_dir:
            self.data_handler.save_file(file_dir)

    def open_file(self):
        file_dir = self.file_browser.browse_file('hdf5 file, (*.hdf5)')
        if file_dir:
            self.data_handler.open_file(file_dir)
            data_structure = self.data_handler.get_info()
            # Add new data set to the list in the GUI
            for data_set_type in data_structure:
                for ds in data_structure[data_set_type]:
                    self.add_data_set_to_list(data_set_type, ds)

            # Get the ROI count (from the first data set)
            if 'data_sets' in data_structure:
                ds = data_structure['data_sets'][0]
                self.data_handler.roi_count = self.data_handler.get_roi_count(ds)

    def new_file(self):
        dlg = QMessageBox(self.gui)
        dlg.setWindowTitle('New Session')
        dlg.setText(f'Are you sure you want start a new session. All unsaved data will be lost!')
        dlg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        dlg.setIcon(QMessageBox.Icon.Question)
        button = dlg.exec()
        if button == QMessageBox.StandardButton.Yes:
            self.data_handler.new_file()
            self.gui.data_sets_list.clear()

    def _create_short_cuts(self):
        pass

    def _connect_short_cuts(self, connect=True):
        pass

    # ==================================================================================================================
    # MOUSE AND KEY PRESS HANDLING
    # ------------------------------------------------------------------------------------------------------------------
    def on_mouse_click(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            pos = event.scenePos()
            mouse_point = self.data_plotter.master_plot.vb.mapSceneToView(pos)
            # Get x value (corresponding to time axis)
            self.mouse_x_pos = mouse_point.x()
            modifiers = QApplication.keyboardModifiers()
            if modifiers == Qt.KeyboardModifier.ControlModifier:  # Ctrl key
                # Enter Cutout Mode
                if self.signal_selection_status:
                    self.signal_selection_status = False
                    self.draw_selection(status='exit')
                else:
                    self.draw_selection()
                    self.signal_selection_status = np.invert(self.signal_selection_status)

    def on_key_press(self, event):
        if event.key() == Qt.Key.Key_Left:
            # left arrow key
            self.prev_roi()
        elif event.key() == Qt.Key.Key_Right:
            # right arrow key
            self.next_roi()
        elif event.key() == Qt.Key.Key_Return and self.signal_selection_status:
            self.cut_selection()

    def closeEvent(self, event):
        retval = self.gui.exit_dialog()

        if retval == QMessageBox.StandardButton.Save:
            # Save before exit
            self.save_file()
            event.accept()
            # if self.peak_detection is not None:
            #     self.peak_detection.main_window_closing.emit()
            # self._save_file()
            self.data_handler.create_new_temp_hdf5_file()
        elif retval == QMessageBox.StandardButton.Discard:
            # Do not save before exit
            event.accept()
            # if self.peak_detection is not None:
            #     self.peak_detection.main_window_closing.emit()
            self.data_handler.create_new_temp_hdf5_file()
        else:
            # Do not exit
            event.ignore()

    # def exit_app(self):
    #     self.data_handler.create_new_temp_hdf5_file()
    #     self.gui.close()

    def mouse_moved(self, event):
        vb = self.gui.trace_plot_item.vb
        if self.gui.trace_plot_item.sceneBoundingRect().contains(event):
            mouse_point = vb.mapSceneToView(event)
            self.gui.mouse_label.setText(f"<p style='color:black'>X： {mouse_point.x():.4f} <br> Y: {mouse_point.y():.4f}</p>")



from roibaview.plugins.base import BasePlugin
from roibaview.gui_utils import DynamicInputDialog
from PyQt6.QtWidgets import QDialog, QMessageBox
import pandas as pd
import os
from pathlib import Path


class ConvertCSVFile(BasePlugin):
    name = "Convert CSV file"
    category = 'utils'

    def __init__(self, **kwargs):
        pass

    def apply(self):
        fields = {
            'input_file': ('', 'file'),
            'save_dir': ('', 'dir'),
            'input_delimiter': ('tab', 'string'),
            'output_delimiter': (',', 'string')
        }

        dialog = DynamicInputDialog(title="Convert CSV File", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()

            # Validate input file
            if not os.path.isfile(params['input_file']):
                QMessageBox.critical(None, "File Error", "The selected input file is not valid.")
                return

            # Handle delimiters
            sep_input = '\t' if params['input_delimiter'] == 'tab' else params['input_delimiter']
            sep_output = '\t' if params['output_delimiter'] == 'tab' else params['output_delimiter']

            try:
                # Read the file
                input_file = pd.read_csv(params['input_file'], sep=sep_input, engine='python')

                # Save the file
                output_path = params['save_dir']
                # output_path = os.path.join(params['save_dir'], "converted.csv")
                input_file.to_csv(output_path, sep=sep_output, index=False)
                QMessageBox.information(None, "Success", f"File saved to:\n{output_path}")

            except Exception as e:
                QMessageBox.critical(None, "Conversion Error", f"An error occurred:\n{str(e)}")


class RemoveColFromCSV(BasePlugin):
    name = "Remove Column from CSV"
    category = 'utils'

    def __init__(self, **kwargs):
        pass

    def apply(self):
        fields = {
            'input_file': ('', 'file'),
            'save_dir': ('', 'dir'),
            'column': (0, 'int'),
            'headers': (True, 'bool')
        }

        dialog = DynamicInputDialog(title="Remove Column", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()

            # Validate input file
            if not os.path.isfile(params['input_file']):
                QMessageBox.critical(None, "File Error", "The selected input file is not valid.")
                return

            # Read the file
            input_file = pd.read_csv(params['input_file'])
            col_nr = params['column']

            # Remove the column
            df = input_file.drop(input_file.columns[col_nr], axis=1)

            # Save to HDD
            save_dir = params['save_dir']
            if params['headers']:
                df.to_csv(save_dir, index=False)
            else:
                df.to_csv(save_dir, index=False, header=None)



from PyQt6.QtWidgets import QMenu


class CustomViewBoxMenu(QMenu):
    def __init__(self, viewbox, parent=None):
        super().__init__(parent)
        self.viewbox = viewbox
        self.setTitle("Custom ViewBox Menu")

        self.addAction("Action 1", self.action1_triggered)
        self.addAction("Action 2", self.action2_triggered)

    def action1_triggered(self):
        print("Action 1 triggered")

    def action2_triggered(self):
        print("Action 2 triggered")



from PyQt6.QtCore import pyqtSignal, Qt
from PyQt6.QtWidgets import QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QFileDialog, QLabel, QCheckBox,\
    QComboBox, QApplication, QDoubleSpinBox
from roibaview.gui import SimpleInputDialog
import ffmpy
import os
import subprocess
import cv2
from tifffile import imwrite
from concurrent.futures import ThreadPoolExecutor
import math
import pandas as pd
from roibaview.plugins.base import BasePlugin


class VideoConverterPlugin(BasePlugin):
    name = "Video Converter"
    category = "tool"

    def __init__(self, config=None, parent=None):
        self.config = config
        self.parent = parent

    def apply(self, *_):
        self.window = VideoConverter(self.config)
        self.window.show()


class VideoConverter(QMainWindow):

    ffmpeg_dir_set = pyqtSignal()

    # Using subprocess to run a terminal command using a string
    # Or using the ffmpeg package
    def __init__(self, convert_settings):
        super().__init__()

        self.settings = convert_settings
        # self.settings.loc['FFMPEG'] = 'NaN'
        self.input_file = None
        self.output_file = None

        self.crf_value = 17
        self.preset = 'superfast'
        self.output_frame_rate = 0

        # Create a central widget and layout
        central_widget = QWidget()
        layout = QVBoxLayout(central_widget)

        # Create a button for browsing files
        self.browse_button = QPushButton("Browse Files")
        self.browse_button.clicked.connect(self.browse_files)

        self.input_file_label = QLabel('Input File not selected')
        self.output_file_label = QLabel('Output File not selected')

        self.start_button = QPushButton("Convert Video")
        self.start_button.clicked.connect(self.start_converting)

        self.extract_frames_button = QPushButton("Extract Frames")
        self.extract_frames_button.clicked.connect(self.start_extracting_frames)

        self.rescale_button = QPushButton("Rescale Video")
        self.rescale_button.clicked.connect(self.rescale_video)

        self.batch_button = QPushButton("Batch Processing ...")
        self.batch_button.clicked.connect(self.batch_processing)

        self.change_ffmpeg_dir_button = QPushButton("Set ffmpeg directory")
        self.change_ffmpeg_dir_button.clicked.connect(self.browse_file_ffmpeg)

        self.ffmpeg_dir_label = QLabel(f'ffmpeg at: {self.settings["FFMPEG"]["dir"]}')

        # Settings
        self.use_gpu = False
        gpu_check_box_layout = QHBoxLayout()
        self.gpu_check_box = QCheckBox()
        self.gpu_check_box.setCheckState(Qt.CheckState.Unchecked)
        self.gpu_check_box.stateChanged.connect(self.get_gpu_state)
        self.gpu_check_box_label = QLabel('Use GPU')
        gpu_check_box_layout.addWidget(self.gpu_check_box_label)
        gpu_check_box_layout.addWidget(self.gpu_check_box)

        supress_terminal_output_layout = QHBoxLayout()
        self.supress_terminal_output = True
        self.supress_terminal_output_check_box = QCheckBox()
        self.supress_terminal_output_check_box.setCheckState(Qt.CheckState.Checked)
        self.supress_terminal_output_check_box.stateChanged.connect(self.get_supress_state)
        self.supress_terminal_output_label = QLabel('Supress Terminal Output')
        supress_terminal_output_layout.addWidget(self.supress_terminal_output_label)
        supress_terminal_output_layout.addWidget(self.supress_terminal_output_check_box)

        self.quality_combo_box = QComboBox()
        self.quality_combo_box.addItem('veryslow')
        self.quality_combo_box.addItem('slower')
        self.quality_combo_box.addItem('slow')
        self.quality_combo_box.addItem('medium')
        self.quality_combo_box.addItem('fast')
        self.quality_combo_box.addItem('faster')
        self.quality_combo_box.addItem('veryfast')
        self.quality_combo_box_label = QLabel('Compression Preset')

        self.constant_rate_factor = QDoubleSpinBox()
        self.constant_rate_factor.setValue(self.crf_value)
        self.constant_rate_factor_label = QLabel('Video Quality: CRF (visually lossless=17, technically lossless=0)(range: 0-51)')

        self.change_frame_rate = QDoubleSpinBox()
        self.change_frame_rate.setValue(0)
        self.change_frame_rate_label = QLabel('Output Frame Rate (Hz) (will be ignored if set to 0)')

        self.status_label = QLabel('Ready')

        layout.addWidget(self.browse_button)
        layout.addWidget(self.input_file_label)
        layout.addWidget(self.output_file_label)
        layout.addWidget(self.change_ffmpeg_dir_button)
        layout.addSpacing(10)
        layout.addWidget(self.ffmpeg_dir_label)
        layout.addSpacing(20)
        layout.addLayout(gpu_check_box_layout)
        layout.addSpacing(5)
        layout.addLayout(supress_terminal_output_layout)
        layout.addSpacing(10)
        layout.addWidget(self.quality_combo_box_label)
        layout.addSpacing(5)
        layout.addWidget(self.quality_combo_box)
        layout.addSpacing(20)
        layout.addWidget(self.constant_rate_factor_label)
        layout.addSpacing(5)
        layout.addWidget(self.constant_rate_factor)
        layout.addSpacing(20)
        layout.addWidget(self.change_frame_rate_label)
        layout.addSpacing(5)
        layout.addWidget(self.change_frame_rate)
        layout.addSpacing(20)
        layout.addWidget(self.start_button)
        layout.addSpacing(10)
        layout.addWidget(self.extract_frames_button)
        layout.addSpacing(10)
        layout.addWidget(self.rescale_button)
        layout.addSpacing(10)
        layout.addWidget(self.batch_button)
        layout.addSpacing(10)

        layout.addWidget(self.status_label)

        # Set the central widget and window properties
        self.setCentralWidget(central_widget)
        self.setWindowTitle("Video Converter")

        if self.settings['FFMPEG']['dir'] == 'NaN':
            self.browse_file_ffmpeg()
        self.ffmpeg_dir = self.settings['FFMPEG']['dir']
        self.ffmpeg_probe = f'{os.path.split(self.ffmpeg_dir)[0]}/ffprobe.exe'
        self._define_ffmpeg_settings()

    def batch_processing(self):
        input_dir = QFileDialog.getExistingDirectory(self)
        if input_dir:
            # Get Files
            print("")
            print("++++ BATCH PROCESSING ++++")
            print("")
            self.input_file_label.setText("Batch Processing ...")
            file_list = os.listdir(input_dir)
            for f_name in file_list:
                self.input_file = f'{input_dir}/{f_name}'
                self.output_file = f'{input_dir}/{f_name[:-4]}_batch.mp4'
                print(f"++++ STARTING: {self.input_file} ++++")
                self._define_ffmpeg_settings()
                self.please_wait_status()
                QApplication.processEvents()
                self.convert_video(self.input_file, self.output_file)
                self.finished_status()

    def convert_to_tiff_stack(self):
        input_video = self.input_file
        output_tiff = self.output_file
        chunk_size = 300

        cap = cv2.VideoCapture(input_video)
        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        num_chunks = math.ceil(total_frames / chunk_size)

        for chunk_idx, start_frame in enumerate(range(0, total_frames, chunk_size)):
            frames = []
            cap_chunk = cv2.VideoCapture(input_video)
            cap_chunk.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
            for _ in range(chunk_size):
                ret, frame = cap_chunk.read()
                if ret:
                    frames.append(frame)
                else:
                    break

            imwrite(output_tiff, frames, append=True, bigtiff=True, compression='deflate')
            print(f"Processed chunk {chunk_idx + 1}/{num_chunks}", end='\r')

            cap_chunk.release()  # Release the video capture object

        cap.release()  # Release the main video capture object

    def convert_to_tiff_stack2(self):
        input_video = self.input_file
        output_tiff = self.output_file
        chunk_size = 300
        cap = cv2.VideoCapture(input_video)
        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        for start_frame in range(0, total_frames, chunk_size):
            end_frame = min(start_frame + chunk_size, total_frames)
            frames = []

            for i in range(start_frame, end_frame):
                cap.set(cv2.CAP_PROP_POS_FRAMES, i)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
                else:
                    break

            imwrite(output_tiff, frames, append=True, bigtiff=True, compression='lzw')
            print(f'{end_frame} of {total_frames} Frames', end='\r')

        cap.release()

    def rescale_video(self):
        size_dialog = SimpleInputDialog('Video Size', 'Video Size (widthxheight): ')
        if size_dialog.exec() == size_dialog.DialogCode.Accepted:
            video_size = size_dialog.get_input()
        else:
            return None
        video_size = video_size.split('x')
        if len(video_size) == 2:
            print(f'New Video Size: {video_size[0]} x {video_size[1]} pixel')
        else:
            print('Input incorrect!')
            return None

        self._define_ffmpeg_settings()
        file_dir_name = os.path.split(self.input_file)
        output_file = file_dir_name[0] + f'/{file_dir_name[1][:-4]}_rescaled.{file_dir_name[1][-3:]}'
        print(f'Store to: {output_file}')

        self.get_supress_state()
        if self.supress_terminal_output:
            global_settings = self.ffmpeg_global_opt['supress']
        else:
            global_settings = self.ffmpeg_global_opt['show']
        ff = ffmpy.FFmpeg(
            executable=self.ffmpeg_dir,
            global_options=global_settings,
            inputs={self.input_file: None},
            # outputs={output_file: ['-vf', 'scale=640:480']}
            outputs={output_file: ['-c:v', 'libx264', '-preset', self.preset, '-crf', str(self.crf_value), '-vf', f'scale={video_size[0]}:{video_size[1]}']}
        )
        ff.run()

    def _define_ffmpeg_settings(self):
        self.ffmpeg_input_opt = {'gpu': ['-hwaccel', 'cuda', '-hwaccel_output_format', 'cuda'], 'cpu': None}

        # Get Values
        self.crf_value = int(self.constant_rate_factor.value())
        self.preset = self.quality_combo_box.currentText()
        self.output_frame_rate = int(self.change_frame_rate.value())

        if self.crf_value > 51:
            self.crf_value = 51

        if self.output_frame_rate > 0:
            self.ffmpeg_output_opt = {
                'gpu': ['-c:v', 'h264_nvenc', '-preset', self.preset, '-qp', str(self.crf_value), '-filter:v', f'fps={self.output_frame_rate}'],
                'cpu': ['-c:v', 'libx264', '-preset', self.preset, '-crf', str(self.crf_value), '-filter:v', f'fps={self.output_frame_rate}'],
                'avi': ['-an', '-vcodec', 'rawvideo', '-y'],
            }
        else:
            self.ffmpeg_output_opt = {
                'gpu': ['-c:v', 'h264_nvenc', '-preset', self.preset, '-qp', str(self.crf_value)],
                'cpu': ['-c:v', 'libx264', '-preset', self.preset, '-crf', str(self.crf_value)],
                'avi': ['-an', '-vcodec', 'rawvideo', '-y'],
            }

        self.ffmpeg_global_opt = {
            'supress': ['-y', '-loglevel', 'quiet'],
            'show': ['-y'],
        }

    def get_gpu_state(self):
        if self.gpu_check_box.checkState() == Qt.CheckState.Checked:
            self.use_gpu = True
            # change presets
            self.quality_combo_box.clear()
            self.quality_combo_box.addItem('slow')
            self.quality_combo_box.addItem('medium')
            self.quality_combo_box.addItem('fast')
        else:
            self.use_gpu = False
            # change presets
            self.quality_combo_box.clear()
            self.quality_combo_box.addItem('veryslow')
            self.quality_combo_box.addItem('slower')
            self.quality_combo_box.addItem('slow')
            self.quality_combo_box.addItem('medium')
            self.quality_combo_box.addItem('fast')
            self.quality_combo_box.addItem('faster')
            self.quality_combo_box.addItem('veryfast')

    def get_supress_state(self):
        if self.supress_terminal_output_check_box.checkState() == Qt.CheckState.Checked:
            self.supress_terminal_output = True
        else:
            self.supress_terminal_output = False

    def get_video_info(self, filename):
        result = subprocess.run([self.ffmpeg_probe, "-v", "error", "-select_streams", "v:0", "-show_entries",
                                 "stream=duration:stream=avg_frame_rate", "-of",
                                 "default=noprint_wrappers=1:nokey=1", filename],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        output = result.stdout.decode("utf-8").split('\r\n')
        frame_rate = float(output[0].split('/')[0])  # Extracting the first part before '/'
        duration = float(output[1])

        return frame_rate, duration

    def extract_frames(self, input_file, save_dir):
        if self.ffmpeg_dir is not None:
            # ffmpeg -i input.mp4 -vf fps=1 %04d.png
            video_frame_rate, video_duration = self.get_video_info(input_file)

            if self.output_frame_rate > 0:
                output_cmd = ['-vf', f'fps={self.output_frame_rate}']
                fr = self.output_frame_rate
            else:
                output_cmd = None
                fr = video_frame_rate

            number_of_frames = int(fr * video_duration)
            print(f'++++ Expecting to Store {number_of_frames} Frames to HDD+++')
            counter = len(str(number_of_frames)) + 1
            output_dir = f'{save_dir}/%0{counter}d.jpg'
            if self.supress_terminal_output:
                global_settings = self.ffmpeg_global_opt['supress']
            else:
                global_settings = self.ffmpeg_global_opt['show']

            ff = ffmpy.FFmpeg(
                executable=self.ffmpeg_dir,
                global_options=global_settings,
                inputs={input_file: None},
                outputs={output_dir: output_cmd}
            )
            ff.run()

    def start_extracting_frames(self):
        file_dir = str(QFileDialog.getExistingDirectory(self, "Select Directory"))
        if self.input_file is not None and file_dir is not None:
            print('START EXTRACTING')
            # self._define_ffmpeg_settings()
            self.output_frame_rate = int(self.change_frame_rate.value())
            self.please_wait_status()
            QApplication.processEvents()
            self.extract_frames(self.input_file, file_dir)
            self.finished_status()

    def convert_video(self, input_file, output_file):
        if self.ffmpeg_dir is not None:
            if output_file[-3:] == 'tif':
                print('CONVERT TO TIFF')
                self.convert_to_tiff_stack()
            else:
                # check settings
                if self.use_gpu:
                    hw = 'gpu'
                else:
                    hw = 'cpu'

                if output_file[-3:] == 'avi':
                    # Use no compression for avi file (otherwise you can not open it in imagej)
                    input_cmd = self.ffmpeg_input_opt['cpu']
                    output_cmd = self.ffmpeg_output_opt['avi']
                else:
                    input_cmd = self.ffmpeg_input_opt[hw]
                    output_cmd = self.ffmpeg_output_opt[hw]

                if self.supress_terminal_output:
                    global_settings = self.ffmpeg_global_opt['supress']
                else:
                    global_settings = self.ffmpeg_global_opt['show']

                ff = ffmpy.FFmpeg(
                    executable=self.ffmpeg_dir,
                    global_options=global_settings,
                    inputs={input_file: input_cmd},
                    outputs={output_file: output_cmd}
                )
                ff.run()

    def browse_file_ffmpeg(self):
        self.ffmpeg_dir, _ = QFileDialog.getOpenFileName(self, "Select FFMPEG .exe", "", "ffmpeg (*.exe)")
        # self.settings.modify_setting('ffmpeg', self.ffmpeg_dir)
        self.settings['FFMPEG']['dir'] = self.ffmpeg_dir
        with open('roibaview/config.ini', 'w') as configfile:
            self.settings.write(configfile)
        self.ffmpeg_dir_label.setText(f'ffmpeg at: {self.ffmpeg_dir}')

    def please_wait_status(self):
        self.status_label.setText('Please wait ... ')
        self.browse_button.setDisabled(True)
        self.change_ffmpeg_dir_button.setDisabled(True)

    def finished_status(self):
        self.status_label.setText('Converting finished!')
        self.browse_button.setDisabled(False)
        self.change_ffmpeg_dir_button.setDisabled(False)
        if not self.supress_terminal_output:
            print("")
            print(f"++++ FINISHED: {os.path.split(self.input_file)[1]} --> {os.path.split(self.output_file)[1]} ++++")
            print("")

    def browse_files(self):
        input_file, _ = QFileDialog.getOpenFileName(
            self, "Select Input File", "", "Video Files (*.mp4; *.avi; *.mkv; *.mpeg; *.mpg)")
        if input_file:
            self.input_file = input_file
            self.input_file_label.setText(input_file)

    def start_converting(self):
        output_file, _ = QFileDialog.getSaveFileName(
            self, "Select Output File", "", "MP4, (*.mp4);; AVI, (*.avi);; MKV, (*.mkv);; TIF, (*.tif)")
        if output_file:
            self.output_file = output_file
            self.output_file_label.setText(output_file)

        if self.input_file is not None and self.output_file is not None:
            self._define_ffmpeg_settings()
            self.please_wait_status()
            QApplication.processEvents()
            self.convert_video(self.input_file, self.output_file)
            self.finished_status()



import h5py
import os
import shutil
import numpy as np
import pandas as pd
import tempfile
from PyQt6.QtCore import pyqtSignal, QObject
from scipy import signal
from roibaview.gui import MessageBox
from scipy.signal import decimate, resample

"""
Notes:
    - Try to use PyTables instead of h5py
    - Try using NixPiy

Data Structure:
.
├── data_sets
│   ├── set_01
│   ├── set_02
:   :
│   └── set_n
│
├── global_data_sets
│   ├── global_set_01
:   :
│   └── global_set_n
│
└── some_other_stuff
    ├── stuff_01
    └── info
"""


class DataHandler(QObject):
    signal_roi_id_changed = pyqtSignal()

    def __init__(self):
        QObject.__init__(self)
        self._set_csv_import_settings()
        temp_dir = tempfile.gettempdir()
        self.temp_file_name = os.path.join(temp_dir, "temp_data.hdf5")
        self.create_new_temp_hdf5_file()
        self.roi_count = 0

    def _set_csv_import_settings(self):
        # Settings for importing a csv file using pandas
        # The decimal symbol (english: ".", german: ",")
        self.csv_decimal = '.'

        # The separation/delimiter symbol: "," or "\t" or ";" or etc.
        self.csv_sep = ','

    def create_new_temp_hdf5_file(self):
        # from IPython import embed
        # embed()
        # exit()
        # Will create an empty hdf5 file into the temp directory with one group called "data_sets"
        with h5py.File(self.temp_file_name, 'w') as f:
            f.create_group('data_sets')
            f.create_group('global_data_sets')

    def import_csv(self, file_dir, data_name, sampling_rate, data_set_type):
        # The .csv file: Each Column is the data of one ROI so the shape is (Samples, ROIs)
        # First check if there are headers (ROI Names)
        data_check = pd.read_csv(file_dir, decimal=self.csv_decimal, sep=self.csv_sep, index_col=None, nrows=1, header=None)
        if data_check.iloc[0, :].dtype == 'O':
            # There are headers
            data_file = pd.read_csv(file_dir, decimal=self.csv_decimal, sep=self.csv_sep, index_col=None)

            # Remove Index Column if there is one
            if 'Unnamed: 0' in data_file.keys():
                data_file.drop(columns='Unnamed: 0', inplace=True)
            headers = np.array(data_file.keys())
            # headers = data_check.iloc[0, :].to_numpy()

        else:
            data_file = pd.read_csv(file_dir, decimal=self.csv_decimal, sep=self.csv_sep, index_col=None, header=None)

            # Remove Index Column if there is one
            if 'Unnamed: 0' in data_file.keys():
                data_file.drop(columns='Unnamed: 0', inplace=True)

            headers = np.arange(0, data_file.shape[1], 1)

        # This needs to be converted to a numpy matrix
        data = data_file.to_numpy()

        # Check dimensions (must match hdf5 style)
        if data.ndim == 1:
            data = np.atleast_2d(data)

        # Open the temp hdf5 file and store data set there
        check = self.add_new_data_set(
            data_set_type, data_name, data,
            sampling_rate=sampling_rate,
            time_offset=0,
            y_offset=0,
            header=headers)

    def get_info(self):
        with h5py.File(self.temp_file_name, 'r') as f:
            groups = list(f.keys())
            results = dict()
            for gr in groups:
                results[gr] = list(f[gr].keys())
        return results

    def check_if_exists(self, data_set_type, data_set_name):
        with h5py.File(self.temp_file_name, 'r') as f:
            if data_set_name in f[data_set_type]:
                print('ERROR: Data set with this name already exists!')
                return True
            else:
                return False

    def delete_column(self, data_set_type, data_set_name, col_nr):
        with h5py.File(self.temp_file_name, 'r+') as f:
            if data_set_name in f[data_set_type]:
                # Convert to a NumPy array
                dset = f[data_set_type][data_set_name]
                data = dset[:]

                # Remove the column
                try:
                    modified_data = np.delete(data, col_nr, axis=1)

                    # Resize the dataset to match new shape
                    dset.resize(modified_data.shape)

                    # Overwrite dataset with new data
                    dset[...] = modified_data  # Overwrite without deleting the dataset
                    self.roi_count = modified_data.shape[1]

                except IndexError:
                    return None

    def delete_data_set(self, data_set_type, data_set_name):
        with h5py.File(self.temp_file_name, 'r+') as f:
            if data_set_name in f[data_set_type]:
                # print('Delete:')
                # print(f[data_set_type][data_set_name])
                # f[data_set_type][data_set_name][:] = 0
                del f[data_set_type][data_set_name]

    def rename_data_set(self, data_set_type, data_set_name, new_name):
        with h5py.File(self.temp_file_name, 'r+') as f:
            if data_set_name in f[data_set_type]:
                f[data_set_type][new_name] = f[data_set_type][data_set_name]
                del f[data_set_type][data_set_name]

    def add_new_data_set(self, data_set_type, data_set_name, data, sampling_rate, time_offset, y_offset, header):
        # Open the temp hdf5 file and store data set there
        already_exists = False
        with h5py.File(self.temp_file_name, 'r+') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                MessageBox(title='ERROR', text='Data set with this name already exists!')
                data_set_name = data_set_name + '_new'
                already_exists = True

            # CREATE NEW DATASET
            new_entry = f[data_set_type].create_dataset(data_set_name, data=data, chunks=True)
            if data_set_type == 'global_data_sets':
                header_name = 'header_names'
            else:
                header_name = 'roi_names'
                self.roi_count = data.shape[1]
            new_entry.attrs[header_name] = header
            new_entry.attrs['sampling_rate'] = float(sampling_rate)
            new_entry.attrs['time_offset'] = time_offset
            new_entry.attrs['y_offset'] = y_offset
            new_entry.attrs['color'] = '#000000'  # black
            new_entry.attrs['lw'] = 1
            new_entry.attrs['name'] = data_set_name
            new_entry.attrs['data_type'] = data_set_type

            return already_exists

    def add_meta_data(self, data_set_type, data_set_name, metadata_dict):
        with h5py.File(self.temp_file_name, 'r+') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                for k in metadata_dict:
                    data_set = f[data_set_type][data_set_name]
                    data_set.attrs[k] = metadata_dict[k]
            else:
                print('ERROR: Data set not found!')

    def get_data_set_meta_data(self, data_set_type, data_set_name):
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                data_set = f[data_set_type][data_set_name]
                meta_data = dict(data_set.attrs)
                return meta_data
            else:
                print('ERROR: Data set not found!')
                return None

    def get_data_set(self, data_set_type, data_set_name):
        # Get a specific data set
        # If it is available store it into a numpy array (RAM) and return it
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f[data_set_type]:
                data_set = f[data_set_type][data_set_name][:]
                return data_set
            else:
                print('ERROR: Data set not found!')
                return None

    def get_roi_data(self, data_set_name, roi_idx):
        # Get the data for a specific ROI in a specific data set
        # If it is available store it into a numpy array (RAM) and return it
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f['data_sets']:
                data_set = f['data_sets'][data_set_name]
            else:
                print('ERROR: Data set not found!')
                return None
            # Check if roi idx is in data set
            if data_set.shape[1] > roi_idx:
                roi_data = data_set[:, roi_idx]
                return roi_data
            else:
                print('ERROR: ROI Index is outside of data set range!')
                return None

    def save_file(self, file_dir):
        shutil.copyfile(self.temp_file_name, file_dir)

    def open_file(self, file_dir):
        shutil.copyfile(file_dir, self.temp_file_name)

    def new_file(self):
        self.create_new_temp_hdf5_file()

    def get_roi_count(self, data_set_name):
        with h5py.File(self.temp_file_name, 'r') as f:
            # Check if data set is available
            if data_set_name in f['data_sets']:
                # Cols = Rois
                roi_count = f['data_sets'][data_set_name].shape[1]
                return roi_count
            else:
                print('ERROR: Data set not found!')
                return None

    @staticmethod
    def compute_time_axis(data_size, fr):
        max_time = data_size / fr
        return np.linspace(0, max_time, data_size)


class TransformData(QObject):
    signal_data_transformed = pyqtSignal()

    def __init__(self):
        QObject.__init__(self)

    @staticmethod
    def down_sampling(data, ds_factor, fs):
        # New sampling rate
        new_fs = fs / ds_factor
        # new_samples = int(data.shape[0] / ds_factor)
        # down_sampled_data = resample(data, new_samples)
        # Apply decimate function to downsample the data
        down_sampled_data = decimate(data, ds_factor, axis=0)

        return down_sampled_data, new_fs

    # @staticmethod
    # def to_z_score(data):
    #     """ Compute Z-Score = (Data - Mean) / SD
    #
    #     :param data: numpy array (columns: ROIs, rows: data points over time)
    #     :return: Data set with z-scored values
    #     """
    #     # Check if there is only one ROI (Column)
    #     if data.shape[1] == 1:
    #         data = data.flatten()
    #     return (data - np.mean(data, axis=0)) / np.std(data, axis=0)

    # @staticmethod
    # def to_delta_f_over_f(data, fr, fbs_per=5, window=None):
    #     """ Compute delta F over F for raw fluorescence values
    #
    #     :param data: numpy array (columns: ROIs, rows: data points over time)
    #     :param fr: frame rate in seconds
    #     :param fbs_per: percentile to calculate baseline (0.0 to 1.0)
    #     :param window: window size in seconds for computing sliding percentile baseline (if None, no window is used)
    #     :return:Delta F over F normalized data set
    #     """
    #     # This is using the pandas rolling method, so we need to convert the data to a DataFrame first
    #     from IPython import embed
    #     df = pd.DataFrame(data)
    #     if window is None:
    #         fbs = np.percentile(df, fbs_per, axis=0)
    #     else:
    #         per_window = int(window * fr)
    #         quant = fbs_per / 100
    #         # fbs = df.rolling(window=per_window, center=True, min_periods=0, axis=0).quantile(quant)
    #         fbs = df.rolling(window=per_window, center=True, min_periods=0).quantile(quant)
    #
    #     df_over_f = (df - fbs) / fbs
    #     return df_over_f.to_numpy()

    # @staticmethod
    # def to_min_max(data):
    #     """ Compute min-max normalization to the range [0, 1]
    #
    #     :param data: numpy array (columns: ROIs, rows: data points over time)
    #     :return: Data normalize to the range [0, 1]
    #     """
    #     return (data - np.min(data, axis=0)) / (np.max(data, axis=0) - np.min(data, axis=0))

    @staticmethod
    def prepare_data(data):
        # Since hdf5 and numpy like their data structured differently, we have to transform it
        # Check if there is only one ROI (Column)
        if data.shape[1] == 1:
            new_data = data.flatten()
        else:
            # otherwise transpose it
            new_data = data.T
        return new_data



from roibaview.plugins.base import BasePlugin
# from roibaview.gui import SimpleInputDialog
from roibaview.gui_utils import DynamicInputDialog
from PyQt6.QtWidgets import QMessageBox, QDialog
import numpy as np
from scipy.signal import butter, sosfiltfilt

"""
# Basic Structure of a Filter Plugin Class with Dialog Input
class FilterPluginName(BasePlugin):
    name = "Filter Name"
    category = "filter"
    
    def __init__(self, **kwargs):
        self.parent = kwargs.get("parent", None)  # needed for dialog parent
    
    # Every Plugin gets the selected data array (samples, rois) and its sampling rate
    def apply(self, data, sampling_rate): 
        dialog = SimpleInputDialog("Filter", "Enter value:", default_value="1")
        # Handling the Dialog:
        if dialog.exec() == dialog.DialogCode.Accepted:
            try:
                threshold = float(dialog.get_input())
            except ValueError:
                QMessageBox.warning(self.parent, "Invalid Input", "Please enter a numeric threshold.")
                return None
        else:
            return None  # Cancelled
        
        # Here comes the code of whatever the Filter is doing
        filter_output = ...
        return filter_output
"""


class ZapFilterPrompt(BasePlugin):
    name = "Zap Filter"
    category = "filter"

    def __init__(self, **kwargs):
        self.parent = kwargs.get("parent", None)  # needed for dialog parent

    def apply(self, data, sampling_rate):
        fields = {
            'threshold': (1, 'float'),
        }

        dialog = DynamicInputDialog(title="Zap Filter", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()
            threshold = params['threshold']
            # Apply threshold-based filtering
            return np.where(data > threshold, 0, data)


class MovingAverageFilter(BasePlugin):
    name = "Moving Average Filter"
    category = "filter"

    # def __init__(self, **kwargs):
    #     pass  # if there is no dialog

    def __init__(self, **kwargs):
        self.parent = kwargs.get("parent", None)  # needed for dialog parent

    def apply(self, data, sampling_rate):
        fields = {
            'window': (100, 'float'),
        }

        dialog = DynamicInputDialog(title="Moving Average Filter", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()
            # (data, fr, window)
            window = params['window']
            window_size = int(window * sampling_rate)
            # Make sure window size is odd
            if window_size % 2 == 0:
                window_size += 1

            pad_width = window_size // 2

            # Define the kernel
            kernel = np.ones(window_size) / window_size
            filtered_data = np.zeros_like(data)
            for i in range(data.shape[1]):
                trace = data[:, i]
                # Pad the input array symmetrically
                padded_array = np.pad(trace, pad_width, mode='symmetric')
                filtered_data[:, i] = np.convolve(padded_array, kernel, mode='valid')

            return filtered_data


class DifferentiateFilter(BasePlugin):
    name = "Differentiate Filter"
    category = "filter"

    def __init__(self, **kwargs):
        pass

    def apply(self, data, sampling_rate):
        return np.diff(data, append=0, axis=0)


class ButterFilterDesign:
    def __init__(self, filter_type, cutoff, fs, order):
        self.filter_type = filter_type
        self.cutoff = cutoff
        self.fs = fs
        self.order = order

    def butter_filter_design(self):
        nyq = 0.5 * self.fs
        normal_cutoff = self.cutoff / nyq
        if normal_cutoff >= 1:
            normal_cutoff = 0.9999
        sos = butter(self.order, normal_cutoff, btype=self.filter_type, output='sos')
        return sos


class LowPassFilter(BasePlugin):
    name = "Low Pass Filter"
    category = 'filter'

    def __init__(self, **kwargs):
        self.parent = kwargs.get("parent", None)  # needed for dialog parent

    def apply(self, data, sampling_rate):
        fields = {
            'cut_off': (100, 'float'),
        }

        dialog = DynamicInputDialog(title="Low Pass Filter", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()
            cut_off = params['cut_off']
            sos = ButterFilterDesign('lowpass', cut_off, sampling_rate, order=2).butter_filter_design()
            return sosfiltfilt(sos, data, axis=0)


class HighPassFilter(BasePlugin):
    name = "High Pass Filter"
    category = 'filter'

    def __init__(self, **kwargs):
        self.parent = kwargs.get("parent", None)  # needed for dialog parent

    def apply(self, data, sampling_rate):
        fields = {
            'cut_off': (100, 'float'),
        }

        dialog = DynamicInputDialog(title="Low Pass Filter", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()
            cut_off = params['cut_off']

            sos = ButterFilterDesign('highpass', cut_off, sampling_rate, order=2).butter_filter_design()
            return sosfiltfilt(sos, data, axis=0)


class EnvelopeFilter(BasePlugin):
    name = "Envelope Filter"
    category = 'filter'

    def __init__(self, **kwargs):
        self.parent = kwargs.get("parent", None)  # needed for dialog parent

    def apply(self, data, sampling_rate):
        fields = {
            'cut_off': (100, 'float'),
        }

        dialog = DynamicInputDialog(title="Low Pass Filter", fields=fields)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            params = dialog.get_inputs()
            cut_off = params['cut_off']

            # Low pass filter the absolute values of the signal in both forward and reverse directions,
            # resulting in zero-phase filtering.
            sos = ButterFilterDesign('lowpass', cut_off, sampling_rate, order=2).butter_filter_design()
            return (np.sqrt(2) * sosfiltfilt(sos, np.abs(data), axis=0)) ** 2
